import os
from unittest.mock import Mock, ANY

import numpy as np
from ezyrb import POD, RBF
from pytest import raises

from pydmd import DMD, ParametricDMD
from pydmd.dmdbase import DMDTimeDict

testdir = "tests/test_datasets/param_dmd/"

# def f1(x,t):
#    return 1./np.cosh(x+3)*np.exp(2.3j*t)
# def f2(x,t):
#    return 2./np.cosh(x)*np.tanh(x)*np.exp(2.8j*t)
# def f(a):
#    def fnc(x,t):
#        return a*f1(x,t) + (1-a)*f2(x,t)
#    return fnc

params = np.linspace(0, 1, 10)

# x = np.linspace(0,np.pi,1000)
# t = np.linspace(0,2*np.pi,100)
# xgrid, tgrid = np.meshgrid(x, t)

# training dataset
# training_data = np.vstack([f(a)(xgrid, tgrid)[None,:] for a in params])
training_data = np.swapaxes(np.load(testdir + "/training_data.npy"), 1, 2)

# test dataset
test_parameters = [0.15, 0.75, 0.28]
# testing_data = np.vstack([f(a)(xgrid, tgrid)[None,:] for a in test_parameters])
testing_data = np.swapaxes(np.load(testdir + "testing_data.npy"), 1, 2)


def test_is_partitioned_1():
    assert not ParametricDMD(None, None, None).is_partitioned


def test_is_partitioned_2():
    assert not ParametricDMD(DMD(), None, None).is_partitioned


def test_is_partitioned_3():
    assert ParametricDMD([DMD() for _ in range(3)], None, None).is_partitioned


def test_init():
    d = DMD()
    p = ParametricDMD(d, sum, np.mean)
    assert p._dmd == d
    assert p._spatial_pod == sum
    assert p._approximation == np.mean


def test_reference_dmd_1():
    d = DMD()
    p = ParametricDMD(d, sum, np.mean)
    assert p._reference_dmd == d


def test_reference_dmd_2():
    l = [DMD() for _ in range(3)]
    p = ParametricDMD(l, sum, np.mean)
    assert p._reference_dmd == l[0]


def test_dmd_time_1():
    d = DMD()
    d.fit(np.ones((10, 100)))
    d.dmd_time["tend"] = 200
    d.dmd_time["t0"] = 100

    p = ParametricDMD(d, sum, np.mean)
    assert p.dmd_time["tend"] == 200
    assert p.dmd_time["t0"] == 100


def test_dmd_time_2():
    l = [DMD() for _ in range(3)]
    d = l[0]
    d.fit(np.ones((10, 100)))
    d.dmd_time["tend"] = 200
    d.dmd_time["t0"] = 100

    p = ParametricDMD(l, sum, np.mean)
    assert p.dmd_time["tend"] == 200
    assert p.dmd_time["t0"] == 100


def test_dmd_timesteps_1():
    d = DMD()
    d.fit(np.ones((10, 100)))
    d.dmd_time["tend"] = 200

    p = ParametricDMD(d, sum, np.mean)
    assert len(p.dmd_timesteps) == 201
    assert p.dmd_timesteps[-1] == 200


def test_dmd_timesteps_2():
    l = [DMD() for _ in range(3)]
    d = l[0]
    d.fit(np.ones((10, 100)))
    d.dmd_time["tend"] = 200

    p = ParametricDMD(l, sum, np.mean)
    assert len(p.dmd_timesteps) == 201
    assert p.dmd_timesteps[-1] == 200


def test_original_time_1():
    d = DMD()
    d.fit(np.ones((10, 100)))
    d.dmd_time["tend"] = 200

    p = ParametricDMD(d, sum, np.mean)
    assert p.original_time["tend"] == 99


def test_original_time_2():
    l = [DMD() for _ in range(3)]
    d = l[0]
    d.fit(np.ones((10, 100)))
    d.dmd_time["tend"] = 200

    p = ParametricDMD(l, sum, np.mean)
    assert p.original_time["tend"] == 99


def test_original_timesteps_1():
    d = DMD()
    d.fit(np.ones((10, 100)))
    d.dmd_time["tend"] = 200

    p = ParametricDMD(d, sum, np.mean)
    assert len(p.original_timesteps) == 100
    assert p.original_timesteps[-1] == 99


def test_original_timesteps_2():
    l = [DMD() for _ in range(3)]
    d = l[0]
    d.fit(np.ones((10, 100)))
    d.dmd_time["tend"] = 200

    p = ParametricDMD(l, sum, np.mean)
    assert len(p.original_timesteps) == 100
    assert p.original_timesteps[-1] == 99


def test_training_parameters_1():
    p = ParametricDMD(None, None, None)
    p._set_training_parameters(np.ones(10))
    assert p.training_parameters.shape == (10, 1)


def test_training_parameters_2():
    p = ParametricDMD(None, None, None)
    p._set_training_parameters([1 for _ in range(10)])
    assert p.training_parameters.shape == (10, 1)


def test_training_parameters_3():
    p = ParametricDMD(None, None, None)
    with raises(ValueError):
        p._set_training_parameters(np.ones((10, 2, 2)))


def test_parameters_1():
    p = ParametricDMD(None, None, None)
    p.parameters = np.ones(10)
    assert p.parameters.shape == (10, 1)


def test_parameters_2():
    p = ParametricDMD(None, None, None)
    p.parameters = [1 for _ in range(10)]
    assert p.parameters.shape == (10, 1)


def test_parameters_3():
    p = ParametricDMD(None, None, None)
    with raises(ValueError):
        p.parameters = np.ones((10, 2, 2))


def test_fit_wrong_training_shape():
    p = ParametricDMD(None, None, None)
    with raises(ValueError):
        p.fit(np.ones((9, 10, 10)), [0 for _ in range(10)])


# assert that fit sets properly _ntrain, _time_instants, _space_dim
def test_fit_sets_quantities():
    p = ParametricDMD(DMD(), POD(), None)
    p.fit(np.ones((10, 9, 8)), [0 for _ in range(10)])

    assert p._ntrain == 10
    assert p._time_instants == 8
    assert p._space_dim == 9


def test_fit_stores_training_parameters():
    p = ParametricDMD(DMD(), POD(), None)
    p.fit(np.ones((10, 9, 8)), [i for i in range(10)])
    assert np.all(p.training_parameters == np.arange(10)[:, None])


def test_fit_stores_2d_training_parameters():
    params = np.hstack(
        [np.arange(10)[:, None], np.flip(np.arange(10))[:, None]]
    )

    p = ParametricDMD(DMD(), POD(), None)
    p.fit(np.ones((10, 9, 8)), params)

    # test shape
    assert p.training_parameters.shape == (10, 2)
    # test that the order is fine
    assert np.all(np.sum(p.training_parameters, axis=1) == np.repeat(9, 10))


def test_compute_training_modal_coefficients_shape():
    p = ParametricDMD(None, POD(rank=5), None)
    p._ntrain = 10

    res = p._compute_training_modal_coefficients(np.ones((50, 60)))
    assert len(res) == 10
    assert res[0].shape == (5, 6)


def test_compute_training_modal_coefficients():
    m = np.vander(np.arange(50) / 50, 60)
    p = ParametricDMD(None, POD(rank=5), None)
    p._ntrain = 10
    np.testing.assert_allclose(
        np.hstack(p._compute_training_modal_coefficients(m)),
        POD(rank=5).fit(m).reduce(m),
    )


def test_compute_training_modal_coefficients2():
    p = ParametricDMD(DMD(svd_rank=-1), POD(rank=5), RBF())
    p.fit(training_data, params)

    inp = np.load(testdir + "space_timemu.npy")
    actual = p._compute_training_modal_coefficients(inp)
    expected = np.load(testdir + "traning_modal_coefficients.npy")

    np.testing.assert_allclose(actual, expected, atol=1.0e-12, rtol=0)


def test_arrange_parametric_snapshots_shape():
    p = ParametricDMD(None, None, None)
    assert p._arrange_parametric_snapshots(np.ones((10, 20, 30))).shape == (
        20,
        300,
    )


def test_arrange_parametric_snapshots():
    p = ParametricDMD(None, None, None)

    m1 = np.array([[1, 2, 3], [4, 5, 6]])
    m2 = np.array(
        [
            [
                0,
                1,
                0,
            ],
            [1, 1, 0],
        ]
    )
    m = np.stack([m1, m2])

    expected = np.array(
        [
            [1, 2, 3, 0, 1, 0],
            [4, 5, 6, 1, 1, 0],
        ]
    )

    np.testing.assert_array_equal(expected, p._arrange_parametric_snapshots(m))


def test_arrange_parametric_snapshots2():
    p = ParametricDMD(DMD(svd_rank=-1), POD(rank=5), RBF())
    p.fit(training_data, params)

    expected = np.load(testdir + "space_timemu.npy")
    actual = p._arrange_parametric_snapshots(training_data)
    np.testing.assert_allclose(actual, expected)


def test_predict_modal_coefficients_shape():
    p = ParametricDMD(DMD(svd_rank=5), POD(rank=10), RBF())
    p.fit(np.ones((10, 20, 40)), np.arange(10))
    p.dmd_time["tend"] = 29
    assert p._predict_modal_coefficients().shape == (10 * 10, 30)


def test_predict_modal_coefficients_partitioned_shape():
    p = ParametricDMD([DMD(svd_rank=5) for _ in range(10)], POD(rank=10), RBF())
    p.fit(np.ones((10, 20, 40)), np.arange(10))
    p.dmd_time["tend"] = 29
    assert p._predict_modal_coefficients().shape == (10 * 10, 30)


def test_predict_modal_coefficients():
    p = ParametricDMD(DMD(svd_rank=-1), POD(rank=5), RBF())
    p.fit(training_data, params)

    expected = np.load(testdir + "forecasted.npy")
    np.testing.assert_allclose(
        p._predict_modal_coefficients(), expected, rtol=0, atol=1.0e-11
    )


def test_interpolate_missing_modal_coefficients_shape():
    p = ParametricDMD(DMD(svd_rank=5), POD(rank=10), RBF())
    p.fit(np.ones((10, 20, 40)), np.arange(10))
    p.dmd_time["tend"] = 49
    p.parameters = [1.5, 5.5, 7.5]
    assert p._interpolate_missing_modal_coefficients(
        np.random.rand(10 * 5, 40)
    ).shape == (3, 5, 40)


# def test_interpolate_missing_modal_coefficients():
#     p = ParametricDMD(DMD(svd_rank=-1), POD(rank=5), RBF())
#     p.fit(training_data, params)
#     p.parameters = test_parameters

#     expected = back_roll_shape(np.load(testdir+'interpolated.npy'))
#     input = np.load(testdir+'forecasted.npy')
#     np.testing.assert_allclose(
#         p._interpolate_missing_modal_coefficients(input),
#         expected, atol=1.e-12, rtol=0)


def reconstructed_data_shape():
    p = ParametricDMD(DMD(svd_rank=5), POD(rank=10), RBF())
    p.fit(np.random.rand(10, 20, 50), np.arange(10))
    p.dmd_time["tend"] = 39
    p.parameters = [1.5, 5.5, 7.5, 8.5]
    assert p.reconstructed_data.shape == (4, 40, 50)


def test_reconstructed_data_noprediction():
    p = ParametricDMD(DMD(svd_rank=-1), POD(rank=5), RBF())
    p.fit(training_data, params)
    p.parameters = test_parameters

    rec = p.reconstructed_data

    np.testing.assert_allclose(rec.real, testing_data.real, atol=1.0e-2)


def test_save():
    p = ParametricDMD(DMD(svd_rank=-1), POD(rank=5), RBF())
    p.fit(training_data, params)
    p.parameters = test_parameters
    p.save("pydmd.test")
    os.remove("pydmd.test")


def test_load():
    p = ParametricDMD(DMD(svd_rank=-1), POD(rank=5), RBF())
    p.fit(training_data, params)
    p.parameters = test_parameters
    p.save("pydmd.test2")
    loaded_p = ParametricDMD.load("pydmd.test2")
    np.testing.assert_array_equal(
        p.reconstructed_data, loaded_p.reconstructed_data
    )
    os.remove("pydmd.test2")


def test_load2():
    p = ParametricDMD(DMD(svd_rank=-1), POD(rank=5), RBF())
    p.fit(training_data, params)
    p.parameters = test_parameters
    p.save("pydmd.test2")
    loaded_p = ParametricDMD.load("pydmd.test2")
    assert isinstance(loaded_p, ParametricDMD)
    os.remove("pydmd.test2")


def test_set_time_monolithic():
    p = ParametricDMD(DMD(svd_rank=-1), POD(rank=5), RBF())
    p.fit(training_data, params)
    p.parameters = test_parameters

    dc = DMDTimeDict()
    dc["t0"] = 10
    dc["tend"] = 20
    dc["dt"] = 1
    p.dmd_time = dc

    assert p.dmd_time == dc
    assert p.dmd_time["t0"] == 10
    assert p.dmd_time["dt"] == 1
    assert p.dmd_time["tend"] == 20


def test_set_time_partitioned():
    dmds = [DMD(svd_rank=-1) for _ in range(len(params))]
    p = ParametricDMD(dmds, POD(rank=5), RBF())
    p.fit(training_data, params)
    p.parameters = test_parameters

    dc = DMDTimeDict()
    dc["t0"] = 10
    dc["tend"] = 20
    dc["dt"] = 1
    p.dmd_time = dc
    p._predict_modal_coefficients()

    for dmd in dmds:
        assert dmd.dmd_time == dc
        assert dmd.dmd_time["t0"] == 10
        assert dmd.dmd_time["dt"] == 1
        assert dmd.dmd_time["tend"] == 20


def test_forecast():
    dmds = [DMD(svd_rank=-1) for _ in range(len(params))]
    p = ParametricDMD(dmds, POD(rank=5), RBF())
    p.fit(training_data, params)
    p.parameters = test_parameters

    dc = DMDTimeDict()
    dc["t0"] = 101
    dc["tend"] = 200
    dc["dt"] = 1
    p.dmd_time = dc

    for r in p.reconstructed_data:
        assert r.shape == (training_data.shape[1], 100)


def test_training_modal_coefficients_property_partitioned():
    dmds = [DMD(svd_rank=-1) for _ in range(len(params))]
    p = ParametricDMD(dmds, POD(rank=5), RBF())
    p.fit(training_data, params)
    assert p.training_modal_coefficients.shape == (10, 5, 100)


def test_training_modal_coefficients_property_monolithic():
    dmd = DMD(svd_rank=-1)
    p = ParametricDMD(dmd, POD(rank=5), RBF())
    p.fit(training_data, params)
    assert p.training_modal_coefficients.shape == (10, 5, 100)


def test_no_training_modal_coefficients_light():
    dmds = [DMD(svd_rank=-1) for _ in range(len(params))]
    p = ParametricDMD(dmds, POD(rank=5), RBF(), light=True)
    p.fit(training_data, params)
    with raises(RuntimeError):
        p.training_modal_coefficients


def test_no_training_modal_coefficients_before_fit():
    dmds = [DMD(svd_rank=-1) for _ in range(len(params))]
    p = ParametricDMD(dmds, POD(rank=5), RBF(), light=True)
    with raises(RuntimeError):
        p.training_modal_coefficients


def test_forecasted_modal_coefficients_shape():
    dmds = [DMD(svd_rank=-1) for _ in range(len(params))]
    p = ParametricDMD(dmds, POD(rank=5), RBF(), light=True)
    p.fit(training_data, params)
    assert p.forecasted_modal_coefficients.shape == (10, 5, 100)
    p.dmd_time["tend"] += 10
    assert p.forecasted_modal_coefficients.shape == (10, 5, 110)


def test_interpolated_modal_coefficients_shape():
    dmds = [DMD(svd_rank=-1) for _ in range(len(params))]
    p = ParametricDMD(dmds, POD(rank=4), RBF(), light=True)
    p.fit(training_data, params)
    p.parameters = [0.25, 0.98, 0.99]
    assert p.interpolated_modal_coefficients.shape == (3, 4, 100)
    p.dmd_time["tend"] += 100
    assert p.interpolated_modal_coefficients.shape == (3, 4, 200)


def test_forecasted_modal_coefficients_reshape():
    dmds = [DMD(svd_rank=-1) for _ in range(len(params))]
    p = ParametricDMD(dmds, POD(rank=4), RBF(), light=True)
    p.fit(training_data, params)
    p.dmd_time["tend"] += 100

    forecasted_modal_coefficients = p._predict_modal_coefficients()
    np.testing.assert_allclose(
        p.forecasted_modal_coefficients[0, 3], forecasted_modal_coefficients[3]
    )
    np.testing.assert_allclose(
        p.forecasted_modal_coefficients[5, 0], forecasted_modal_coefficients[20]
    )


def test_interpolated_modal_coefficients_reshape():
    dmds = [DMD(svd_rank=-1) for _ in range(len(params))]
    p = ParametricDMD(dmds, POD(rank=5), RBF(), light=True)
    p.fit(training_data, params)
    p.dmd_time["tend"] += 100
    p.parameters = [0.25, 0.98, 0.99]

    forecasted_modal_coefficients = p._predict_modal_coefficients()
    interpolated_modal_coefficients = p._interpolate_missing_modal_coefficients(
        forecasted_modal_coefficients
    )
    np.testing.assert_allclose(
        p.interpolated_modal_coefficients[1, 3],
        interpolated_modal_coefficients[1, 3],
    )
    np.testing.assert_allclose(
        p.interpolated_modal_coefficients[2, 0],
        interpolated_modal_coefficients[2, 0],
    )


def test_forecasted_modal_coefficients_shape_monolithic():
    dmds = DMD(svd_rank=-1)
    p = ParametricDMD(dmds, POD(rank=5), RBF(), light=True)
    p.fit(training_data, params)
    assert p.forecasted_modal_coefficients.shape == (10, 5, 100)
    p.dmd_time["tend"] += 10
    assert p.forecasted_modal_coefficients.shape == (10, 5, 110)


def test_interpolated_modal_coefficients_shape_monolithic():
    dmds = DMD(svd_rank=-1)
    p = ParametricDMD(dmds, POD(rank=4), RBF(), light=True)
    p.fit(training_data, params)
    p.parameters = [0.25, 0.98, 0.99]
    assert p.interpolated_modal_coefficients.shape == (3, 4, 100)
    p.dmd_time["tend"] += 100
    assert p.interpolated_modal_coefficients.shape == (3, 4, 200)


def test_forecasted_modal_coefficients_reshape_monolithic():
    dmds = DMD(svd_rank=-1)
    p = ParametricDMD(dmds, POD(rank=4), RBF(), light=True)
    p.fit(training_data, params)
    p.dmd_time["tend"] += 100

    forecasted_modal_coefficients = p._predict_modal_coefficients()
    np.testing.assert_allclose(
        p.forecasted_modal_coefficients[0, 3], forecasted_modal_coefficients[3]
    )
    np.testing.assert_allclose(
        p.forecasted_modal_coefficients[5, 0], forecasted_modal_coefficients[20]
    )


def test_interpolated_modal_coefficients_reshape_monolithic():
    dmds = DMD(svd_rank=-1)
    p = ParametricDMD(dmds, POD(rank=5), RBF(), light=True)
    p.fit(training_data, params)
    p.dmd_time["tend"] += 100
    p.parameters = [0.25, 0.98, 0.99]

    forecasted_modal_coefficients = p._predict_modal_coefficients()
    interpolated_modal_coefficients = p._interpolate_missing_modal_coefficients(
        forecasted_modal_coefficients
    )
    np.testing.assert_allclose(
        p.interpolated_modal_coefficients[1, 3],
        interpolated_modal_coefficients[1, 3],
    )
    np.testing.assert_allclose(
        p.interpolated_modal_coefficients[2, 0],
        interpolated_modal_coefficients[2, 0],
    )


def test_no_parameters_error():
    dmds = DMD(svd_rank=-1)
    p = ParametricDMD(dmds, POD(rank=5), RBF(), light=True)
    p.fit(training_data, params)
    p.dmd_time["tend"] += 100

    with raises(
        ValueError,
        match="Unknown parameters not found. Did you set `ParametricDMD.parameters`?",
    ):
        p.reconstructed_data


def init_sub_dmd():
    dmd = Mock()
    dmd.fit = Mock()
    return dmd


def test_dmd_fit_args_monolithic():
    dmd = init_sub_dmd()
    p = ParametricDMD(
        dmd,
        POD(rank=5),
        RBF(),
        dmd_fit_args=(1, 2, None),
        dmd_fit_kwargs={"label": "PIPPO", "n": 2},
    )
    p.fit(training_data, params)
    dmd.fit.assert_called_once_with(ANY, 1, 2, None, label="PIPPO", n=2)


def test_dmd_fit_args_partitioned():
    dmds = [init_sub_dmd() for _ in range(len(training_data))]
    p = ParametricDMD(
        dmds,
        POD(rank=5),
        RBF(),
        dmd_fit_args=(1, 2, None),
        dmd_fit_kwargs={"label": "PIPPO", "n": 2},
    )
    p.fit(training_data, params)
    for dmd in dmds:
        dmd.fit.assert_called_once_with(ANY, 1, 2, None, label="PIPPO", n=2)



"""
Zero-mean pre-processing.
"""

import sys
from typing import Tuple

import numpy as np

from pydmd.dmdbase import DMDBase
from pydmd.preprocessing.pre_post_processing import (
    PrePostProcessing,
    PrePostProcessingDMD,
)

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override


def zero_mean_preprocessing(dmd: DMDBase, *args, **kwargs):
    """
    Zero-mean pre-processing. All exceeding arguments are passed to
    `np.average`.

    :param dmd: DMD instance to be wrapped.
    """

    return PrePostProcessingDMD(
        dmd, _ZeroMeanPrePostProcessing(*args, **kwargs)
    )


class _ZeroMeanPrePostProcessing(PrePostProcessing):
    def __init__(self, *args, **kwargs):
        self._average_args = args
        self._average_kwargs = kwargs

    @override
    def pre_processing(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        if "keepdims" not in self._average_kwargs:
            self._average_kwargs["keepdims"] = True
        mean = np.average(X, *self._average_args, **self._average_kwargs)
        return mean, X - mean

    @override
    def post_processing(
        self, pre_processing_output: np.ndarray, Y: np.ndarray
    ) -> np.ndarray:
        return Y + pre_processing_output



"""
Pre/post-processing capability for DMD instances.
"""

from __future__ import annotations

from inspect import isroutine
from typing import Any, Dict, Generic, Tuple, TypeVar

import numpy as np

from pydmd.dmdbase import DMDBase

# Pre-processing output type
S = TypeVar("S")


class PrePostProcessing(Generic[S]):
    def pre_processing(self, X: np.ndarray) -> Tuple[S, np.ndarray]:
        return None, X

    def post_processing(
        self, pre_processing_output: S, Y: np.ndarray
    ) -> np.ndarray:
        return Y


class PrePostProcessingDMD(Generic[S]):
    """
    Pre/post-processing decorator. This class is not thread-safe in case of
    stateful transformations.

    :param dmd: DMD instance to be decorated.
    :param pre_processing: Pre-processing function, receives a state holder
        `dict` for stateful preprocessing, and positional/keyword arguments
        passed to`DMDBase.fit()`. The returned values from this function are
        passed exactly in the same order to the wrapped DMD instance.
    :param post_processing: Post-processing function, receives the state
        holder created during the pre-processing and the value of the
        reconstructed data from the wrapped DMD instance.
    """

    def __init__(
        self,
        dmd: DMDBase,
        pre_post_processing: PrePostProcessing[S] = PrePostProcessing(),
    ):
        if dmd is None:
            raise ValueError("DMD instance cannot be None")
        self._pre_post_processing = pre_post_processing

        self._dmd = dmd
        self._pre_processing_output: S | None = None

    def __getattribute__(self, name):
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            pass

        if "fit" == name:
            return self._pre_processing_fit

        if "reconstructed_data" == name:
            return self._reconstructed_data_with_post_processing()

        # This check is needed to allow copy/deepcopy
        if name != "_dmd":
            sub_dmd = self._dmd
            if isinstance(sub_dmd, PrePostProcessingDMD):
                return PrePostProcessingDMD.__getattribute__(sub_dmd, name)
            return object.__getattribute__(self._dmd, name)
        return None

    @property
    def pre_post_processed_dmd(self):
        """
        Return the pre/post-processed DMD instance.

        :return: decorated DMD instance.
        :rtype: pydmd.DMDBase
        """
        return self._dmd

    @property
    def modes_activation_bitmask(self):
        return self._dmd.modes_activation_bitmask

    @modes_activation_bitmask.setter
    def modes_activation_bitmask(self, value):
        self._dmd.modes_activation_bitmask = value

    def _pre_processing_fit(self, *args, **kwargs):
        X = PrePostProcessingDMD._extract_training_data(*args, **kwargs)
        self._pre_processing_output, pre_processed_training_data = (
            self._pre_post_processing.pre_processing(X)
        )
        new_args, new_kwargs = PrePostProcessingDMD._replace_training_data(
            pre_processed_training_data, *args, **kwargs
        )
        return self._dmd.fit(*new_args, **new_kwargs)

    def _reconstructed_data_with_post_processing(self) -> np.ndarray:
        data = self._dmd.reconstructed_data

        if not isroutine(data):
            return self._pre_post_processing.post_processing(
                self._pre_processing_output,
                data,
            )

        # e.g. DMDc
        def output(*args, **kwargs) -> np.ndarray:
            return self._pre_post_processing.post_processing(
                self._pre_processing_output,
                data(*args, **kwargs),
            )

        return output

    @staticmethod
    def _extract_training_data(*args, **kwargs):
        if len(args) >= 1:
            return args[0]
        elif "X" in kwargs:
            return kwargs["X"]
        raise ValueError(
            f"Could not extract training data from {args=}, {kwargs=}"
        )

    @staticmethod
    def _replace_training_data(
        new_training_data: Any, *args, **kwargs
    ) -> [Tuple[Any, ...], Dict[str, Any]]:
        if len(args) >= 1:
            return (new_training_data,) + args[1:], kwargs
        elif "X" in kwargs:
            new_kwargs = dict(kwargs)
            new_kwargs["X"] = new_training_data
            return args, new_kwargs



import numpy as np
import scipy.io
from pytest import raises

from pydmd import DMD, SpDMD

data = np.load("tests/test_datasets/heat_90.npy")
gammas = [1.0e-1, 0.5, 2, 5, 10, 20, 40, 50, 100]


def test_number_nonzero_amplitudes_rho1():
    zeros = np.load("tests/test_datasets/zero_amplitudes_rho1.npy")

    for gm, z in zip(gammas, zeros.T):
        assert all(
            SpDMD(
                svd_rank=30,
                gamma=gm,
                release_memory=False,
            )
            .fit(data)
            ._find_zero_amplitudes()
            == z
        )


def test_number_nonzero_amplitudes_rho1e4():
    zeros = np.load("tests/test_datasets/zero_amplitudes_rho1e4.npy")

    for gm, z in zip(gammas, zeros.T):
        assert all(
            SpDMD(
                svd_rank=30,
                gamma=gm,
                rho=1.0e4,
                release_memory=False,
            )
            .fit(data)
            ._find_zero_amplitudes()
            == z
        )


def test_spdmd_amplitudes_rho1():
    amps = np.load("tests/test_datasets/spdmd_amplitudes_rho1.npy")

    for gm, z in zip(gammas, amps.T):
        np.testing.assert_allclose(
            SpDMD(
                svd_rank=30,
                gamma=gm,
            )
            .fit(data)
            .amplitudes,
            z,
            rtol=2,
            atol=1.0e-6,
        )


def test_spdmd_amplitudes_rho1e4():
    amps = np.load("tests/test_datasets/spdmd_amplitudes_rho1e4.npy")

    for gm, z in zip(gammas, amps.T):
        np.testing.assert_allclose(
            SpDMD(svd_rank=30, gamma=gm, rho=1.0e4).fit(data).amplitudes,
            z,
            rtol=2,
            atol=1.0e-6,
        )


def test_rho():
    assert SpDMD().rho == 1
    assert SpDMD(rho=10).rho == 10


def test_maxiter():
    assert SpDMD()._max_iterations == 10000
    assert SpDMD(max_iterations=2)._max_iterations == 2


def test_gamma():
    assert SpDMD().gamma == 10
    assert SpDMD(gamma=2).gamma == 2


def test_exact():
    assert SpDMD()._exact == True


def test_abstol():
    assert SpDMD()._abs_tol == 1.0e-6
    assert SpDMD(abs_tolerance=1.0e10)._abs_tol == 1.0e10


def test_reltol():
    assert SpDMD()._rel_tol == 1.0e-4
    assert SpDMD(rel_tolerance=1.0e10)._rel_tol == 1.0e10


def test_verbose():
    assert SpDMD()._verbose == True
    assert SpDMD(verbose=False)._verbose == False


def test_enforce_zero():
    assert SpDMD()._enforce_zero == True
    assert SpDMD(enforce_zero=False)._enforce_zero == False


def test_release_memory():
    assert SpDMD()._release_memory == True
    assert SpDMD(release_memory=False)._release_memory == False


def test_zero_tolerance():
    assert SpDMD()._zero_absolute_tolerance == 1.0e-12
    assert (
        SpDMD(zero_absolute_tolerance=1.0e-6)._zero_absolute_tolerance == 1.0e-6
    )


def test_zero_tolerance_no_zeros():
    zero_amps = (
        SpDMD(zero_absolute_tolerance=1.0e30, release_memory=False)
        .fit(data)
        ._find_zero_amplitudes()
    )
    assert all(zero_amps)


def test_zero_tolerance_all_zeros():
    zero_amps = (
        SpDMD(zero_absolute_tolerance=0, release_memory=False)
        .fit(data)
        ._find_zero_amplitudes()
    )
    assert all(np.logical_not(zero_amps))


def test_release_memory_releases():
    o = SpDMD(release_memory=True).fit(data)
    assert o._P is None
    assert o._Plow is None
    assert o._q is None

    o = SpDMD(release_memory=False).fit(data)
    assert o._P is not None
    assert o._Plow is not None
    assert o._q is not None


def test_update_lagrangian():
    alpha = np.random.rand(10)
    beta = np.random.rand(10)
    lmbd = np.random.rand(10)

    np.testing.assert_allclose(
        SpDMD(rho=0)._update_lagrangian(alpha, beta, lmbd), lmbd
    )
    np.testing.assert_allclose(
        SpDMD()._update_lagrangian(alpha, alpha, lmbd), lmbd
    )


def test_update_alpha():
    o = SpDMD(release_memory=False).fit(data)

    beta = np.random.rand(len(o.amplitudes))
    lmbd = np.random.rand(len(o.amplitudes))

    uk = beta - lmbd / o.rho

    rhs = np.linalg.solve(o._Plow, o._q + uk * o.rho / 2)
    lhs = o._Plow.conj().T

    np.testing.assert_allclose(lhs.dot(o._update_alpha(beta, lmbd)), rhs)


def test_get_bitmask_default():
    dmd = SpDMD(release_memory=True, svd_rank=-1)
    dmd.fit(X=data)
    assert np.all(dmd.modes_activation_bitmask == True)


def test_set_bitmask():
    dmd = SpDMD(release_memory=True, svd_rank=-1)
    dmd.fit(X=data)

    new_bitmask = np.full(len(dmd.amplitudes), True, dtype=bool)
    new_bitmask[[0]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.modes_activation_bitmask[0] == False
    assert np.all(dmd.modes_activation_bitmask[1:] == True)


def test_not_fitted_get_bitmask_raises():
    dmd = SpDMD(release_memory=True, svd_rank=-1)
    with raises(RuntimeError):
        print(dmd.modes_activation_bitmask)


def test_not_fitted_set_bitmask_raises():
    dmd = SpDMD(release_memory=True, svd_rank=-1)
    with raises(RuntimeError):
        dmd.modes_activation_bitmask = np.full(3, True, dtype=bool)


def test_raise_wrong_dtype_bitmask():
    dmd = SpDMD(release_memory=True, svd_rank=-1)
    dmd.fit(X=data)
    with raises(RuntimeError):
        dmd.modes_activation_bitmask = np.full(3, 0.1)


def test_fitted():
    dmd = SpDMD(release_memory=True, svd_rank=-1)
    assert not dmd.fitted
    dmd.fit(X=data)
    assert dmd.fitted


def test_bitmask_amplitudes():
    dmd = SpDMD(release_memory=True, svd_rank=-1)
    dmd.fit(X=data)

    old_n_amplitudes = dmd.amplitudes.shape[0]
    retained_amplitudes = np.delete(dmd.amplitudes, [0, -1])

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.amplitudes.shape[0] == old_n_amplitudes - 2
    np.testing.assert_almost_equal(dmd.amplitudes, retained_amplitudes)


def test_bitmask_eigs():
    dmd = SpDMD(release_memory=True, svd_rank=-1)
    dmd.fit(X=data)

    old_n_eigs = dmd.eigs.shape[0]
    retained_eigs = np.delete(dmd.eigs, [0, -1])

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.eigs.shape[0] == old_n_eigs - 2
    np.testing.assert_almost_equal(dmd.eigs, retained_eigs)


def test_bitmask_modes():
    dmd = SpDMD(release_memory=True, svd_rank=-1)
    dmd.fit(X=data)

    old_n_modes = dmd.modes.shape[1]
    retained_modes = np.delete(dmd.modes, [0, -1], axis=1)

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.modes.shape[1] == old_n_modes - 2
    np.testing.assert_almost_equal(dmd.modes, retained_modes)


def test_reconstructed_data():
    dmd = SpDMD(release_memory=True, svd_rank=-1)
    dmd.fit(X=data)

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    dmd.reconstructed_data
    assert True


def test_getitem_modes():
    dmd = SpDMD(release_memory=True, svd_rank=-1)
    dmd.fit(X=data)
    old_n_modes = dmd.modes.shape[1]

    assert dmd[[0, -1]].modes.shape[1] == 2
    np.testing.assert_almost_equal(dmd[[0, -1]].modes, dmd.modes[:, [0, -1]])

    assert dmd.modes.shape[1] == old_n_modes

    assert dmd[1::2].modes.shape[1] == old_n_modes // 2
    np.testing.assert_almost_equal(dmd[1::2].modes, dmd.modes[:, 1::2])

    assert dmd.modes.shape[1] == old_n_modes

    assert dmd[[1, 3]].modes.shape[1] == 2
    np.testing.assert_almost_equal(dmd[[1, 3]].modes, dmd.modes[:, [1, 3]])

    assert dmd.modes.shape[1] == old_n_modes

    assert dmd[2].modes.shape[1] == 1
    np.testing.assert_almost_equal(np.squeeze(dmd[2].modes), dmd.modes[:, 2])

    assert dmd.modes.shape[1] == old_n_modes


def test_getitem_raises():
    dmd = SpDMD(release_memory=True, svd_rank=-1)
    dmd.fit(X=data)

    with raises(ValueError):
        dmd[[0, 1, 1, 0, 1]]
    with raises(ValueError):
        dmd[[True, True, False, True]]
    with raises(ValueError):
        dmd[1.0]


def test_correct_amplitudes():
    dmd = SpDMD(release_memory=True, svd_rank=-1)
    dmd.fit(X=data)
    np.testing.assert_array_almost_equal(dmd.amplitudes, dmd._b)



import numpy as np
from pytest import raises

from pydmd import DMD, FbDMD
from pydmd.mrdmd import MrDMD


def create_data(t_size=1600):
    x = np.linspace(-10, 10, 80)
    t = np.linspace(0, 20, t_size)
    Xm, Tm = np.meshgrid(x, t)

    D = np.exp(-np.power(Xm / 2, 2)) * np.exp(0.8j * Tm)
    D += np.sin(0.9 * Xm) * np.exp(1j * Tm)
    D += np.cos(1.1 * Xm) * np.exp(2j * Tm)
    D += 0.6 * np.sin(1.2 * Xm) * np.exp(3j * Tm)
    D += 0.6 * np.cos(1.3 * Xm) * np.exp(4j * Tm)
    D += 0.2 * np.sin(2.0 * Xm) * np.exp(6j * Tm)
    D += 0.2 * np.cos(2.1 * Xm) * np.exp(8j * Tm)
    D += 0.1 * np.sin(5.7 * Xm) * np.exp(10j * Tm)
    D += 0.1 * np.cos(5.9 * Xm) * np.exp(12j * Tm)
    D += 0.1 * np.random.randn(*Xm.shape)
    D += 0.03 * np.random.randn(*Xm.shape)
    D += (
        5
        * np.exp(-np.power((Xm + 5) / 5, 2))
        * np.exp(-np.power((Tm - 5) / 5, 2))
    )
    D[:800, 40:] += 2
    D[200:600, 50:70] -= 3
    D[800:, :40] -= 2
    D[1000:1400, 10:30] += 3
    D[1000:1080, 50:70] += 2
    D[1160:1240, 50:70] += 2
    D[1320:1400, 50:70] += 2
    return D.T


sample_data = create_data()


def test_max_level_threshold():
    level = 10
    dmd = DMD()
    mrdmd = MrDMD(dmd, max_level=level, max_cycles=2)
    mrdmd.fit(X=sample_data)
    lvl_threshold = int(np.log(sample_data.shape[1] / 4.0) / np.log(2.0)) + 1
    assert lvl_threshold == mrdmd.max_level


def test_partial_time_interval():
    level = 4
    dmd = DMD()
    mrdmd = MrDMD(dmd, max_level=level, max_cycles=2)
    mrdmd.fit(X=sample_data)
    ans = {"t0": 1200, "tend": 1400.0, "delta": 200.0}
    assert mrdmd.partial_time_interval(3, 6) == ans


def test_partial_time_interval2():
    level = 4
    dmd = DMD()
    mrdmd = MrDMD(dmd, max_level=level, max_cycles=2)
    mrdmd.fit(X=sample_data)
    with raises(ValueError):
        mrdmd.partial_time_interval(5, 0)


def test_partial_time_interval3():
    level = 4
    dmd = DMD()
    mrdmd = MrDMD(dmd, max_level=level, max_cycles=2)
    mrdmd.fit(X=sample_data)
    with raises(ValueError):
        mrdmd.partial_time_interval(3, 8)


def test_time_window_bins():
    level = 4
    dmd = DMD()
    mrdmd = MrDMD(dmd, max_level=level, max_cycles=2)
    mrdmd.fit(X=sample_data)
    assert len(mrdmd.time_window_bins(0, 1600)) == 2**5 - 1


def test_time_window_bins2():
    level = 3
    dmd = DMD()
    mrdmd = MrDMD(dmd, max_level=level, max_cycles=2)
    mrdmd.fit(X=sample_data)
    expected_bins = np.array([[0, 0], [1, 0], [2, 0], [2, 1], [3, 1], [3, 2]])
    np.testing.assert_array_equal(
        mrdmd.time_window_bins(200, 600), expected_bins
    )


def test_time_window_eigs():
    level = 2
    dmd = DMD()
    mrdmd = MrDMD(dmd, max_level=level, max_cycles=1)
    mrdmd.fit(X=sample_data)
    exp = sum([len(dmd.eigs) for dmd in mrdmd])
    assert len(mrdmd.time_window_eigs(0, 1600)) == exp


def test_time_window_frequency():
    level = 2
    dmd = DMD()
    mrdmd = MrDMD(dmd, max_level=level, max_cycles=1)
    mrdmd.fit(X=sample_data)
    exp = sum([len(dmd.frequency) for dmd in mrdmd])
    assert len(mrdmd.time_window_frequency(0, 1600)) == exp


def test_time_window_growth_rate():
    level = 2
    dmd = DMD()
    mrdmd = MrDMD(dmd, max_level=level, max_cycles=1)
    mrdmd.fit(X=sample_data)
    exp = sum([len(dmd.growth_rate) for dmd in mrdmd])
    assert len(mrdmd.time_window_growth_rate(0, 1600)) == exp


def test_time_window_amplitudes():
    level = 2
    dmd = DMD()
    mrdmd = MrDMD(dmd, max_level=level, max_cycles=1)
    mrdmd.fit(X=sample_data)
    exp = sum([len(dmd.amplitudes) for dmd in mrdmd])
    assert len(mrdmd.time_window_amplitudes(0, 1600)) == exp


def test_shape_modes():
    level = 2
    dmd = DMD(svd_rank=1)
    mrdmd = MrDMD(dmd, max_level=level, max_cycles=1)
    mrdmd.fit(X=sample_data)
    assert mrdmd.modes.shape == (sample_data.shape[0], 2 ** (level + 1) - 1)


def test_shape_dynamics():
    level = 2
    dmd = DMD(svd_rank=1)
    mrdmd = MrDMD(dmd, max_level=level, max_cycles=1)
    mrdmd.fit(X=sample_data)
    assert mrdmd.dynamics.shape == (2 ** (level + 1) - 1, sample_data.shape[1])


def test_reconstructed_data():
    dmd = DMD(svd_rank=1)
    mrdmd = MrDMD(dmd, max_level=6, max_cycles=2)
    mrdmd.fit(X=sample_data)
    dmd_data = mrdmd.reconstructed_data
    norm_err = np.linalg.norm(sample_data - dmd_data) / np.linalg.norm(
        sample_data
    )
    assert norm_err < 1


def test_partial_modes1():
    max_level = 5
    level = 2
    rank = 2
    dmd = DMD(svd_rank=rank)
    mrdmd = MrDMD(dmd, max_level=6, max_cycles=2)
    mrdmd.fit(X=sample_data)
    pmodes = mrdmd.partial_modes(level)
    assert pmodes.shape == (sample_data.shape[0], 2**level * rank)


def test_partial_modes2():
    max_level = 5
    level = 2
    rank = 2
    dmd = DMD(svd_rank=rank)
    mrdmd = MrDMD(dmd, max_level=6, max_cycles=2)
    mrdmd.fit(X=sample_data)
    pmodes = mrdmd.partial_modes(level, 3)
    assert pmodes.shape == (sample_data.shape[0], rank)


def test_partial_dynamics1():
    max_level = 5
    level = 2
    rank = 2
    dmd = DMD(svd_rank=rank)
    mrdmd = MrDMD(dmd, max_level=6, max_cycles=2)
    mrdmd.fit(X=sample_data)
    pdynamics = mrdmd.partial_dynamics(level)
    assert pdynamics.shape == (2**level * rank, sample_data.shape[1])


def test_partial_dynamics2():
    max_level = 5
    level = 2
    rank = 2
    dmd = DMD(svd_rank=rank)
    mrdmd = MrDMD(dmd, max_level=6, max_cycles=2)
    mrdmd.fit(X=sample_data)
    pdynamics = mrdmd.partial_dynamics(level, 3)
    assert pdynamics.shape == (rank, sample_data.shape[1] // 2**level)


def test_eigs2():
    max_level = 5
    level = 2
    rank = -1
    dmd = DMD(svd_rank=rank)
    mrdmd = MrDMD(dmd, max_level=6, max_cycles=2)
    mrdmd.fit(X=sample_data)
    assert mrdmd.eigs.ndim == 1


def test_partial_eigs1():
    max_level = 5
    level = 2
    rank = 2
    dmd = DMD(svd_rank=rank)
    mrdmd = MrDMD(dmd, max_level=max_level, max_cycles=2)
    mrdmd.fit(X=sample_data)
    peigs = mrdmd.partial_eigs(level)
    assert peigs.shape == (rank * 2**level,)


def test_partial_eigs2():
    max_level = 5
    level = 2
    rank = 2
    dmd = DMD(svd_rank=rank)
    mrdmd = MrDMD(dmd, max_level=6, max_cycles=2)
    mrdmd.fit(X=sample_data)
    peigs = mrdmd.partial_eigs(level, 3)
    assert peigs.shape == (rank,)


def test_partial_eigs1():
    max_level = 5
    level = 2
    rank = 2
    dmd = DMD(svd_rank=rank)
    mrdmd = MrDMD(dmd, max_level=max_level, max_cycles=2)
    mrdmd.fit(X=sample_data)
    pamps = mrdmd.partial_amplitudes(level)
    assert pamps.shape == (rank * 2**level,)


def test_partial_eigs2():
    max_level = 5
    level = 2
    rank = 2
    dmd = DMD(svd_rank=rank)
    mrdmd = MrDMD(dmd, max_level=6, max_cycles=2)
    mrdmd.fit(X=sample_data)
    pamps = mrdmd.partial_amplitudes(level, 3)
    assert pamps.shape == (rank,)


def test_partial_reconstructed1():
    max_level = 5
    level = 2
    rank = 2
    dmd = DMD(svd_rank=rank)
    mrdmd = MrDMD(dmd, max_level=6, max_cycles=2)
    mrdmd.fit(X=sample_data)
    pdata = mrdmd.partial_reconstructed_data(level)
    assert pdata.shape == sample_data.shape


def test_partial_reconstructed2():
    max_level = 5
    level = 2
    rank = 2
    dmd = DMD(svd_rank=rank)
    mrdmd = MrDMD(dmd, max_level=max_level, max_cycles=2)
    mrdmd.fit(X=sample_data)
    pdata = mrdmd.partial_reconstructed_data(level, 3)
    assert pdata.shape == (
        sample_data.shape[0],
        sample_data.shape[1] // 2**level,
    )


def test_wrong_partial_reconstructed():
    max_level = 5
    level = 2
    rank = 2
    dmd = DMD(svd_rank=rank)
    mrdmd = MrDMD(dmd, max_level=max_level, max_cycles=2)
    mrdmd.fit(X=sample_data)
    with raises(ValueError):
        pdata = mrdmd.partial_reconstructed_data(max_level + 1, 2)


def test_wrong_level():
    max_level = 5
    rank = 2
    dmd = DMD(svd_rank=rank)
    mrdmd = MrDMD(dmd, max_level=max_level, max_cycles=2)
    mrdmd.fit(X=sample_data)
    with raises(ValueError):
        mrdmd.partial_modes(max_level + 1)


def test_wrong_bin():
    max_level = 5
    level = 2
    dmd = DMD()
    mrdmd = MrDMD(dmd, max_level=max_level, max_cycles=2)
    mrdmd.fit(X=sample_data)
    with raises(ValueError):
        mrdmd.partial_modes(level=level, node=2**level)


def test_consistency():
    level = 5
    dmd = DMD()
    mrdmd = MrDMD(dmd, max_level=level, max_cycles=1)
    mrdmd.fit(X=sample_data)
    np.testing.assert_array_almost_equal(
        mrdmd.reconstructed_data, mrdmd.modes @ mrdmd.dynamics
    )


def test_consistency2():
    import sys

    import numpy

    numpy.set_printoptions(threshold=sys.maxsize)

    mrdmd = MrDMD(DMD(), max_level=5, max_cycles=1)
    mrdmd.fit(X=create_data(t_size=1400))

    np.testing.assert_array_almost_equal(
        mrdmd.reconstructed_data, mrdmd.modes @ mrdmd.dynamics
    )


def test_bitmask_not_implemented():
    with raises(RuntimeError):
        mrdmd = MrDMD(DMD(), max_level=5, max_cycles=1)
        mrdmd.fit(X=sample_data)
        mrdmd.modes_activation_bitmask
    with raises(RuntimeError):
        mrdmd = MrDMD(DMD(), max_level=5, max_cycles=1)
        mrdmd.fit(X=sample_data)
        mrdmd.modes_activation_bitmask = None


def test_getitem_not_implemented():
    with raises(RuntimeError):
        mrdmd = MrDMD(DMD(), max_level=5, max_cycles=1)
        mrdmd.fit(X=sample_data)
        mrdmd[1:3]


def test_one_dmd():
    m = MrDMD(dmd=DMD(svd_rank=4), max_level=5)

    for level in range(6):
        leaves = 2**level
        for leaf in range(leaves):
            dmd = m.dmd_tree[level, leaf]
            assert isinstance(dmd, DMD)
            assert dmd.operator._svd_rank == 4


def test_list_dmd():
    l = [DMD(svd_rank=5 - i) for i in range(4)]
    m = MrDMD(dmd=l, max_level=3)

    for level in range(4):
        leaves = 2**level
        for leaf in range(leaves):
            dmd = m.dmd_tree[level, leaf]
            assert isinstance(dmd, DMD)
            assert dmd.operator._svd_rank == 5 - level


def test_tuple_dmd():
    l = tuple(DMD(svd_rank=5 - i) for i in range(4))
    m = MrDMD(dmd=l, max_level=3)

    for level in range(4):
        leaves = 2**level
        for leaf in range(leaves):
            dmd = m.dmd_tree[level, leaf]
            assert isinstance(dmd, DMD)
            assert dmd.operator._svd_rank == 5 - level


def test_list_wrong_size_dmd():
    l = [DMD(svd_rank=5 - i) for i in range(4)]
    with raises(ValueError):
        MrDMD(dmd=l, max_level=4)


def test_tuple_dmd():
    l = tuple(DMD(svd_rank=5 - i) for i in range(5))
    with raises(ValueError):
        MrDMD(dmd=l, max_level=3)


def test_func_dmd():
    def f(level, leaf):
        return FbDMD(svd_rank=level * leaf)

    m = MrDMD(dmd=f, max_level=5)

    for level in range(6):
        leaves = 2**level
        for leaf in range(leaves):
            dmd = m.dmd_tree[level, leaf]
            assert isinstance(dmd, FbDMD)
            assert dmd.operator._svd_rank == level * leaf


def test_quantitative_list_dmd():
    l = [DMD(svd_rank=4) for i in range(4)]
    m1 = MrDMD(dmd=l, max_level=3).fit(X=sample_data).reconstructed_data
    m2 = (
        MrDMD(DMD(svd_rank=4), max_level=3)
        .fit(X=sample_data)
        .reconstructed_data
    )
    np.testing.assert_almost_equal(m1, m2)


def test_quantitative_tuple_dmd():
    l = tuple(DMD(svd_rank=4) for i in range(4))
    m1 = MrDMD(dmd=l, max_level=3).fit(X=sample_data).reconstructed_data
    m2 = (
        MrDMD(DMD(svd_rank=4), max_level=3)
        .fit(X=sample_data)
        .reconstructed_data
    )
    np.testing.assert_almost_equal(m1, m2)


def test_quantitative_func_dmd():
    def f(level, leaf):
        return FbDMD(svd_rank=4)

    m1 = MrDMD(dmd=f, max_level=4).fit(sample_data).reconstructed_data
    m2 = (
        MrDMD(dmd=FbDMD(svd_rank=4), max_level=4)
        .fit(sample_data)
        .reconstructed_data
    )



import numpy as np
from pytest import raises, warns
from scipy.integrate import solve_ivp
from numpy.testing import assert_allclose, assert_equal

from pydmd.lando import LANDO

# Chaotic Lorenz parameters:
sigma, rho, beta = 10, 28, 8 / 3

# Lorenz system fixed point:
x_bar = [-np.sqrt(beta * (rho - 1)), -np.sqrt(beta * (rho - 1)), rho - 1]

# True linear operator at the fixed point:
A_true = np.array(
    [
        [-sigma, sigma, 0.0],
        [1.0, -1.0, np.sqrt(beta * (rho - 1))],
        [-np.sqrt(beta * (rho - 1)), -np.sqrt(beta * (rho - 1)), -beta],
    ]
)

# Settings to replicate odeint defaults.
solve_ivp_opts = {}
solve_ivp_opts["rtol"] = 1e-12
solve_ivp_opts["atol"] = 1e-12
solve_ivp_opts["method"] = "LSODA"


def relative_error(mat, mat_true):
    """
    Computes and returns the relative error between two matrices.
    """
    return np.linalg.norm(mat - mat_true) / np.linalg.norm(mat_true)


def differentiate(X_data, time_step):
    """
    Method for performing 2nd order finite difference. Assumes the input
    matrix X is 2-D, with uniformly-sampled snapshots filling each column.
    Requires the time step between each snapshot.
    """
    if not isinstance(X_data, np.ndarray) or X_data.ndim != 2:
        raise ValueError("Please ensure that input data is a 2D array.")
    X_prime = np.empty(X_data.shape)
    X_prime[:, 1:-1] = (X_data[:, 2:] - X_data[:, :-2]) / (2 * time_step)
    X_prime[:, 0] = (-3 * X_data[:, 0] + 4 * X_data[:, 1] - X_data[:, 2]) / (
        2 * time_step
    )
    X_prime[:, -1] = (3 * X_data[:, -1] - 4 * X_data[:, -2] + X_data[:, -3]) / (
        2 * time_step
    )
    return X_prime


def generate_lorenz_data(t_eval, x0=(-8, 8, 27)):
    """
    Given a time vector t_eval = t1, t2, ..., evaluates and
    returns the snapshots of the Lorenz system as columns of
    the matrix X for the initial condition x0.
    """

    def lorenz_system(state_t, state):
        x, y, z = state
        x_dot = sigma * (y - x)
        y_dot = (x * (rho - z)) - y
        z_dot = (x * y) - (beta * z)
        return [x_dot, y_dot, z_dot]

    sol = solve_ivp(
        lorenz_system,
        [t_eval[0], t_eval[-1]],
        x0,
        t_eval=t_eval,
        **solve_ivp_opts,
    )

    return sol.y


def dummy_kernel(Xk, Yk):
    """
    Externally-defined linear kernel function.
    """
    return Xk.T.dot(Yk)


def dummy_grad(Xk, yk):
    """
    Externally-defined linear kernel function gradient.
    """
    return Xk.T.dot(np.eye(len(yk)))


# Generate Lorenz system data.
dt = 0.001
t = np.arange(0, 10, dt)
X = generate_lorenz_data(t)
Y = differentiate(X, dt)

# Generate short version of the data.
X_short = X[:, :2000]
Y_short = Y[:, :2000]

# Set the LANDO learning parameters used by most test models.
lando_params = {}
lando_params["svd_rank"] = 3
lando_params["kernel_metric"] = "poly"
lando_params["kernel_params"] = {"degree": 3, "coef0": 1.0, "gamma": 1.0}
lando_params["x_rescale"] = 1.0 / np.abs(X).max(axis=1)
lando_params["dict_tol"] = 1e-6


def test_fitted():
    """
    Test that the partially fitted and fitted attributes are updated correctly.
    """
    lando = LANDO(**lando_params)
    assert not lando.partially_fitted
    assert not lando.fitted

    lando.fit(X_short, Y_short)
    assert lando.partially_fitted
    assert not lando.fitted

    lando.analyze_fixed_point(x_bar)
    assert lando.partially_fitted
    assert lando.fitted


def test_sparse_dictionary():
    """
    Test that the shapes of the sparse dictionary and that the sparse
    dictionary weights are as expected.
    """
    lando = LANDO(**lando_params)

    with raises(ValueError):
        _ = lando.sparse_dictionary

    lando.fit(X_short, Y_short)
    assert X_short.shape[0] == lando.sparse_dictionary.shape[0]
    assert X_short.shape[-1] > lando.sparse_dictionary.shape[-1]
    assert lando.operator.weights.shape == lando.sparse_dictionary.shape


def test_f():
    """
    Test that the computed function f() is accurate.
    """
    lando = LANDO(**lando_params)
    lando.fit(X, Y)
    assert relative_error(lando.f(X), Y) < 1e-5


def test_bias():
    """
    Test that the computed bias term is accurate.
    """
    lando = LANDO(**lando_params)
    lando.fit(X, Y)
    lando.analyze_fixed_point(x_bar)
    assert lando.bias.shape == (3, 1)
    assert np.linalg.norm(lando.bias) < 1e-3


def test_linear():
    """
    Test that the computed linear operator is accurate.
    """
    lando = LANDO(**lando_params)
    lando.fit(X, Y)
    lando.analyze_fixed_point(x_bar, compute_A=True)

    eigs_true, modes_true = np.linalg.eig(A_true)
    modes_rescale = np.divide(lando.modes[0], modes_true[0])
    modes_true_rescaled = np.multiply(modes_true, modes_rescale)

    assert relative_error(lando.linear, A_true) < 1e-4
    assert_allclose(lando.eigs, eigs_true, rtol=1e-4)
    assert_allclose(lando.modes, modes_true_rescaled, rtol=1e-3)


def test_nonlinear():
    """
    Test that the nonlinear operator is accurate.
    """
    lando = LANDO(**lando_params)
    lando.fit(X, Y)
    lando.analyze_fixed_point(x_bar)
    assert lando.nonlinear(X).shape == X.shape

    X_centered = X - np.array(x_bar)[..., None]
    N_est = lando.nonlinear(X_centered)
    N_true = Y - A_true.dot(X_centered)
    assert relative_error(N_est, N_true) < 1e-4


def test_predict_1():
    """
    Test that predict() can be used to reconstruct the system.
    """
    lando = LANDO(**lando_params)
    lando.fit(X, Y)
    lando_recon = lando.predict(
        x0=(-8, 8, 27),
        tend=len(t),
        continuous=True,
        dt=dt,
        solve_ivp_opts=solve_ivp_opts,
    )
    assert relative_error(lando_recon, X) < 0.01


def test_predict_2():
    """
    Test that predict() can be used to predict the system.
    """
    t_long = np.arange(0, 12, dt)
    X_long = generate_lorenz_data(t_long)

    lando = LANDO(**lando_params)
    lando.fit(X, Y)

    lando_predict = lando.predict(
        x0=(-8, 8, 27),
        tend=len(t_long),
        continuous=True,
        dt=dt,
        solve_ivp_opts=solve_ivp_opts,
    )
    assert relative_error(lando_predict, X_long) < 0.1


def test_predict_3():
    """
    Test that predict() can accurately produce data for new initial conditions.
    """
    new_init = (10, 14, 10)
    X2 = generate_lorenz_data(t, x0=new_init)

    lando = LANDO(**lando_params)
    lando.fit(X, Y)

    lando_predict = lando.predict(
        x0=new_init,
        tend=len(t),
        continuous=True,
        dt=dt,
        solve_ivp_opts=solve_ivp_opts,
    )
    assert relative_error(lando_predict, X2) < 0.01


def test_predict_4():
    """
    Test that predict() works for discrete time.
    """
    lando = LANDO(**lando_params)
    lando.fit(X)
    lando_predict = lando.predict(
        x0=(-8, 8, 27),
        tend=len(t),
        continuous=False,
    )
    assert relative_error(lando_predict, X) < 0.1


def test_online_1():
    """
    Test that a LANDO model fitted with the online option yields the
    same results as a LANDO model fitted without the online option.
    """
    lando = LANDO(**lando_params)
    lando.fit(X, Y)

    lando_online = LANDO(online=True, **lando_params)
    lando_online.fit(X, Y)

    assert relative_error(lando_online.f(X), lando.f(X)) < 1e-3


def test_online_2():
    """
    Test that a LANDO model fitted with the online option and an update yields
    the same results as a LANDO model fitted without the online option.
    """
    lando = LANDO(**lando_params)
    lando.fit(X, Y)

    batch_split = int(0.8 * len(t))
    lando_online = LANDO(online=True, **lando_params)
    lando_online.fit(X[:, :batch_split], Y[:, :batch_split])
    lando_online.update(X[:, batch_split:], Y[:, batch_split:])

    assert relative_error(lando_online.f(X), lando.f(X)) < 1e-3


def test_online_3():
    """
    Test that a LANDO model fitted with the online option and an update
    properly updates fixed point analysis results after the update.
    """
    lando = LANDO(**lando_params)
    lando.fit(X, Y)
    lando.analyze_fixed_point(x_bar, compute_A=True)

    batch_split = int(0.8 * len(t))
    lando_online = LANDO(online=True, **lando_params)
    lando_online.fit(X[:, :batch_split], Y[:, :batch_split])
    lando_online.analyze_fixed_point(x_bar, compute_A=True)
    lando_online.update(X[:, batch_split:], Y[:, batch_split:])

    assert_equal(lando_online.fixed_point, lando.fixed_point)
    assert np.linalg.norm(lando_online.bias) < 1e-3
    assert relative_error(lando_online.linear, lando.linear) < 1e-3
    assert relative_error(lando_online.nonlinear(X), lando.nonlinear(X)) < 1e-3


def test_default_kernel():
    """
    Test that there are no errors when calling the default linear kernel.
    """
    lando = LANDO()
    lando.fit(X_short, Y_short)
    lando.analyze_fixed_point(x_bar)


def test_rbf_kernel():
    """
    Test that there are no errors when calling the default RBF kernel.
    """
    lando = LANDO(kernel_metric="rbf")
    lando.fit(X_short, Y_short)
    lando.analyze_fixed_point(x_bar)


def test_custom_kernel():
    """
    Test that there are no errors when using custom kernel functions.
    """
    lando = LANDO(kernel_function=dummy_kernel, kernel_gradient=dummy_grad)
    lando.fit(X_short, Y_short)
    lando.analyze_fixed_point(x_bar)


def test_custom_kernel_error():
    """
    Test that an error occurs if a user attempts a fixed point analysis with a
    custom kernel function but without a kernel gradient function.
    """
    lando = LANDO(kernel_function=dummy_kernel)
    lando.fit(X_short, Y_short)

    with raises(ValueError):
        lando.analyze_fixed_point(x_bar)


def test_kernel_inputs():
    """
    Tests various errors caught by the test_kernel_inputs function.
    """
    # Error should be thrown if an invalid kernel metric is given.
    with raises(ValueError):
        _ = LANDO(kernel_metric="blah")

    # Error should be thrown if kernel_params isn't a dict.
    with raises(TypeError):
        _ = LANDO(kernel_metric="poly", kernel_params=3)

    # Error should be thrown if kernel_params contains invalid entries.
    with raises(ValueError):
        _ = LANDO(kernel_metric="poly", kernel_params={"blah": 3})


def test_kernel_functions_1():
    """
    Tests various errors caught by the test_kernel_functions function.
    Tests for errors related to invalid inputs and combinations.
    """
    # Warning should arise if a kernel function is given without a gradient.
    with warns():
        _ = LANDO(kernel_function=dummy_kernel)

    # Error should be thrown if a gradient is given without a kernel function.
    with raises(ValueError):
        _ = LANDO(kernel_gradient=dummy_grad)

    # Error should be thrown if kernel_function isn't a function.
    with raises(TypeError):
        _ = LANDO(kernel_function=0, kernel_gradient=dummy_grad)

    # Error should be thrown if kernel_gradient isn't a function.
    with raises(TypeError):
        _ = LANDO(kernel_function=dummy_kernel, kernel_gradient=0)


def test_kernel_functions_2():
    """
    Tests various errors caught by the test_kernel_functions function.
    Tests for errors related to invalid function inputs.
    """

    # Define functions that malfunction when called:
    def bad_kernel_1(Xk, Yk):
        return dummy_kernel(Xk.T, Yk)

    def bad_grad_1(Xk, yk):
        return dummy_grad(Xk.T, yk)

    # Define functions that yield incorrect dimensions:
    def bad_kernel_2(Xk, Yk):
        return dummy_kernel(Xk, Yk).T

    def bad_grad_2(Xk, yk):
        return dummy_grad(Xk, yk).T

    with raises(ValueError):
        _ = LANDO(kernel_function=bad_kernel_1, kernel_gradient=dummy_grad)

    with raises(ValueError):
        _ = LANDO(kernel_function=bad_kernel_2, kernel_gradient=dummy_grad)

    with raises(ValueError):
        _ = LANDO(kernel_function=dummy_kernel, kernel_gradient=bad_grad_1)

    with raises(ValueError):
        _ = LANDO(kernel_function=dummy_kernel, kernel_gradient=bad_grad_2)


def test_supported_kernels():
    """
    Test a call to supported kernels.
    """
    lando = LANDO()
    print(lando.supported_kernels)


def test_errors_f():
    """
    Test that expected errors are thrown when calling f.
    """
    lando = LANDO()

    # Error should be thrown if f is called prior to fitting.
    with raises(ValueError):
        _ = lando.f(X_short)

    lando.fit(X_short, Y_short)

    # Error should be thrown if f is given data with the wrong dimension.
    with raises(ValueError):
        _ = lando.f(X_short[:-1])


def test_errors_predict():
    """
    Test that expected errors are thrown when calling predict.
    """
    lando = LANDO()

    # Error should be thrown if called prior to fitting.
    with raises(ValueError):
        _ = lando.predict(x0=(-8, 8, 27), tend=len(t))

    lando.fit(X_short, Y_short)

    # Error should be thrown if data is the wrong dimension.
    with raises(ValueError):
        _ = lando.predict(x0=(-8, 8), tend=len(t))


def test_errors_fixed_point():
    """
    Test that expected errors are thrown when calling analyze_fixed_point.
    """
    lando = LANDO()

    # Error should be thrown if called prior to fitting.
    with raises(ValueError):
        lando.analyze_fixed_point(x_bar)

    lando.fit(X_short, Y_short)

    # Error should be thrown if fixed point is the wrong dimension.
    with raises(ValueError):
        lando.analyze_fixed_point(x_bar[:-1])


def test_errors_update():
    """
    Test that expected errors are thrown when calling update.
    """
    lando = LANDO()

    # Error should be thrown if called prior to fitting.
    with raises(ValueError):
        lando.update(X_short, Y_short)

    lando.fit(X_short, Y_short)

    # Error should be thrown if data is the wrong dimension.
    with raises(ValueError):
        lando.update(X_short, Y_short[:, :-1])


def test_errors_get():
    """
    Test that errors are thrown if the following are attempted to be retrieved
    prior to fully fitting: fixed_point, bias, linear, nonlinear.
    """
    lando = LANDO()

    # Errors should be thrown if called prior to fitting:
    with raises(ValueError):
        _ = lando.fixed_point

    with raises(ValueError):
        _ = lando.bias

    with raises(ValueError):
        _ = lando.linear

    with raises(ValueError):
        _ = lando.nonlinear(X_short)

    lando.fit(X_short, Y_short)

    # Errors should still be thrown after a call to just fit:
    with raises(ValueError):
        _ = lando.fixed_point

    with raises(ValueError):
        _ = lando.bias

    with raises(ValueError):
        _ = lando.linear

    with raises(ValueError):
        _ = lando.nonlinear(X_short)


def test_reconstructed_data():
    """
    Test reconstructed data shape and output.
    """
    lando = LANDO()
    lando.fit(X_short, Y_short)
    lando.analyze_fixed_point(x_bar)

    # Calling for reconstructed data should yield a warning.
    with warns():
        X_recon = lando.reconstructed_data

    assert X_recon.shape == X_short.shape



from builtins import range

import numpy as np
from pytest import raises

from pydmd.dmd import DMD

# 15 snapshot with 400 data. The matrix is 400x15 and it contains
# the following data: f1 + f2 where
# f1 = lambda x,t: sech(x+3)*(1.*np.exp(1j*2.3*t))
# f2 = lambda x,t: (sech(x)*np.tanh(x))*(2.*np.exp(1j*2.8*t))
sample_data = np.load("tests/test_datasets/input_sample.npy")


def create_noisy_data():
    mu = 0.0
    sigma = 0.0  # noise standard deviation
    m = 100  # number of snapshot
    noise = np.random.normal(mu, sigma, m)  # gaussian noise
    A = np.array([[1.0, 1.0], [-1.0, 2.0]])
    A /= np.sqrt(3)
    n = 2
    X = np.zeros((n, m))
    X[:, 0] = np.array([0.5, 1.0])
    # evolve the system and perturb the data with noise
    for k in range(1, m):
        X[:, k] = A.dot(X[:, k - 1])
        X[:, k - 1] += noise[k - 1]
    return X


noisy_data = create_noisy_data()


def test_shape():
    dmd = DMD(svd_rank=-1)
    dmd.fit(X=sample_data)
    assert dmd.modes.shape[1] == sample_data.shape[1] - 1


def test_truncation_shape():
    dmd = DMD(svd_rank=3)
    dmd.fit(X=sample_data)
    assert dmd.modes.shape[1] == 3


def test_rank():
    dmd = DMD(svd_rank=0.9)
    dmd.fit(X=sample_data)
    assert len(dmd.eigs) == 2


def test_Atilde_shape():
    dmd = DMD(svd_rank=3)
    dmd.fit(X=sample_data)
    assert dmd.operator.as_numpy_array.shape == (
        dmd.operator._svd_rank,
        dmd.operator._svd_rank,
    )


def test_Atilde_values():
    dmd = DMD(svd_rank=2)
    dmd.fit(X=sample_data)
    exact_atilde = np.array(
        [
            [-0.70558526 + 0.67815084j, 0.22914898 + 0.20020143j],
            [0.10459069 + 0.09137814j, -0.57730040 + 0.79022994j],
        ]
    )
    np.testing.assert_allclose(exact_atilde, dmd.operator.as_numpy_array)


def test_eigs_1():
    dmd = DMD(svd_rank=-1)
    dmd.fit(X=sample_data)
    assert len(dmd.eigs) == 14


def test_eigs_2():
    dmd = DMD(svd_rank=5)
    dmd.fit(X=sample_data)
    assert len(dmd.eigs) == 5


def test_eigs_3():
    dmd = DMD(svd_rank=2)
    dmd.fit(X=sample_data)
    expected_eigs = np.array(
        [-8.09016994e-01 + 5.87785252e-01j, -4.73868662e-01 + 8.80595532e-01j]
    )
    np.testing.assert_almost_equal(dmd.eigs, expected_eigs, decimal=6)


def test_dynamics_1():
    dmd = DMD(svd_rank=5)
    dmd.fit(X=sample_data)
    assert dmd.dynamics.shape == (5, sample_data.shape[1])


def test_dynamics_2():
    dmd = DMD(svd_rank=1)
    dmd.fit(X=sample_data)
    expected_dynamics = np.array(
        [
            [
                -2.20639502 - 9.10168802e-16j,
                1.55679980 - 1.49626864e00j,
                -0.08375915 + 2.11149018e00j,
                -1.37280962 - 1.54663768e00j,
                2.01748787 + 1.60312745e-01j,
                -1.53222592 + 1.25504678e00j,
                0.23000498 - 1.92462280e00j,
                1.14289644 + 1.51396355e00j,
                -1.83310653 - 2.93174173e-01j,
                1.49222925 - 1.03626336e00j,
                -0.35015209 + 1.74312867e00j,
                -0.93504202 - 1.46738182e00j,
                1.65485808 + 4.01263449e-01j,
                -1.43976061 + 8.39117825e-01j,
                0.44682540 - 1.56844403e00j,
            ]
        ]
    )
    np.testing.assert_allclose(dmd.dynamics, expected_dynamics)


def test_dynamics_opt_1():
    dmd = DMD(svd_rank=5, opt=True)
    dmd.fit(X=sample_data)
    assert dmd.dynamics.shape == (5, sample_data.shape[1])


def test_dynamics_opt_2():
    dmd = DMD(svd_rank=1, opt=True)
    dmd.fit(X=sample_data)
    expected_dynamics = np.array(
        [
            [
                -4.609718826226513 - 6.344781724790875j,
                7.5552686987577165 + 1.3506997434096375j,
                -6.246864367654589 + 4.170577993207872j,
                1.5794144248628537 - 7.179014663490048j,
                3.754043295828462 + 6.13648812118528j,
                -6.810262177959786 - 1.7840079278093528j,
                6.015047060133346 - 3.35961532862783j,
                -1.9658025630719695 + 6.449604262000736j,
                -2.9867632454837936 - 5.8838563367460734j,
                6.097558230017521 + 2.126086276430128j,
                -5.7441543819530265 + 2.6349291080417103j,
                2.266111252852836 - 5.7545702519088895j,
                2.303531963541068 + 5.597105176945707j,
                -5.421019770795679 - 2.3870927539102658j,
                5.443800581850978 - 1.9919716610066682j,
            ]
        ]
    )
    np.testing.assert_allclose(dmd.dynamics, expected_dynamics)


def test_reconstructed_data():
    dmd = DMD()
    dmd.fit(X=sample_data)
    dmd_data = dmd.reconstructed_data
    np.testing.assert_allclose(dmd_data, sample_data)


def test_original_time():
    dmd = DMD(svd_rank=2)
    dmd.fit(X=sample_data)
    expected_dict = {"dt": 1, "t0": 0, "tend": 14}
    np.testing.assert_equal(dmd.original_time, expected_dict)


def test_original_timesteps():
    dmd = DMD()
    dmd.fit(X=sample_data)
    np.testing.assert_allclose(
        dmd.original_timesteps, np.arange(sample_data.shape[1])
    )


def test_dmd_time_1():
    dmd = DMD(svd_rank=2)
    dmd.fit(X=sample_data)
    expected_dict = {"dt": 1, "t0": 0, "tend": 14}
    np.testing.assert_equal(dmd.dmd_time, expected_dict)


def test_dmd_time_2():
    dmd = DMD()
    dmd.fit(X=sample_data)
    dmd.dmd_time["t0"] = 10
    dmd.dmd_time["tend"] = 14
    expected_data = sample_data[:, -5:]
    np.testing.assert_allclose(dmd.reconstructed_data, expected_data)


def test_dmd_time_3():
    dmd = DMD()
    dmd.fit(X=sample_data)
    dmd.dmd_time["t0"] = 8
    dmd.dmd_time["tend"] = 11
    expected_data = sample_data[:, 8:12]
    np.testing.assert_allclose(dmd.reconstructed_data, expected_data)


def test_dmd_time_4():
    dmd = DMD(svd_rank=3)
    dmd.fit(X=sample_data)
    dmd.dmd_time["t0"] = 20
    dmd.dmd_time["tend"] = 20
    expected_data = np.array(
        [
            [-7.29383297e00 - 4.90248179e-14j],
            [-5.69109796e00 - 2.74068833e00j],
            [3.38410649e-83 + 3.75677740e-83j],
        ]
    )
    np.testing.assert_almost_equal(dmd.dynamics, expected_data, decimal=6)


def test_dmd_time_5():
    """
    Let us check that the timestep is actually reduced and not rounded to
    the closest integer.
    """
    dmd = DMD(svd_rank=3)
    dmd.fit(X=sample_data)
    dmd.dmd_time["t0"] = 5
    dmd.dmd_time["tend"] = 20
    dmd.dmd_time["dt"] = 0.5
    reconstruction = dmd.reconstructed_data
    assert np.linalg.norm(reconstruction[0] - reconstruction[1]) != 0


# we check that modes are the same vector multiplied by a coefficient
# when we rescale
def test_rescale_mode_auto_same_modes():
    dmd_no_rescale = DMD(svd_rank=2, opt=True, rescale_mode=None)
    dmd_no_rescale.fit(X=sample_data)

    dmd_auto_rescale = DMD(svd_rank=2, opt=True, rescale_mode="auto")
    dmd_auto_rescale.fit(X=sample_data)

    def normalize(vector):
        return vector / np.linalg.norm(vector)

    dmd_rescale_normalized_modes = np.apply_along_axis(
        normalize, 0, dmd_auto_rescale.modes
    )
    dmd_no_rescale_normalized_modes = np.apply_along_axis(
        normalize, 0, dmd_no_rescale.modes
    )

    np.testing.assert_almost_equal(
        dmd_no_rescale_normalized_modes, dmd_rescale_normalized_modes, decimal=3
    )


# we check that modes are the same vector multiplied by a coefficient
# when we rescale
def test_rescale_mode_custom_same_modes():
    dmd_no_rescale = DMD(svd_rank=2, opt=True, rescale_mode=None)
    dmd_no_rescale.fit(X=sample_data)

    dmd_rescale = DMD(svd_rank=2, opt=True, rescale_mode=np.linspace(5, 10, 2))
    dmd_rescale.fit(X=sample_data)

    def normalize(vector):
        return vector / np.linalg.norm(vector)

    dmd_rescale_normalized_modes = np.apply_along_axis(
        normalize, 0, dmd_rescale.modes
    )
    dmd_no_rescale_normalized_modes = np.apply_along_axis(
        normalize, 0, dmd_no_rescale.modes
    )

    np.testing.assert_almost_equal(
        dmd_no_rescale_normalized_modes, dmd_rescale_normalized_modes, decimal=3
    )


def test_rescale_mode_same_evolution():
    dmd_no_rescale = DMD(svd_rank=5, opt=True, exact=True, rescale_mode=None)
    dmd_no_rescale.fit(X=sample_data)
    dmd_no_rescale.dmd_time["tend"] *= 2

    dmd_rescale = DMD(
        svd_rank=5, opt=True, exact=True, rescale_mode=np.linspace(5, 10, 5)
    )
    dmd_rescale.fit(X=sample_data)
    dmd_rescale.dmd_time["tend"] *= 2

    np.testing.assert_almost_equal(
        dmd_rescale.reconstructed_data,
        dmd_no_rescale.reconstructed_data,
        decimal=6,
    )


def test_rescale_mode_coefficients_count_check():
    dmd_rescale = DMD(svd_rank=5, opt=True, rescale_mode=np.linspace(5, 10, 6))
    with raises(ValueError):
        dmd_rescale.fit(X=sample_data)


def test_predict():
    def f1(x, t):
        return 1.0 / np.cosh(x + 3) * np.exp(2.3j * t)

    def f2(x, t):
        return 2.0 / np.cosh(x) * np.tanh(x) * np.exp(2.8j * t)

    x = np.linspace(-2, 2, 4)
    t = np.linspace(0, 4 * np.pi, 10)

    xgrid, tgrid = np.meshgrid(x, t)

    X1 = f1(xgrid, tgrid)
    X2 = f2(xgrid, tgrid)
    X = X1 + X2

    dmd = DMD()
    dmd.fit(X.T)

    expected = np.array(
        [
            [
                0.35407111 + 0.31966903j,
                0.0581077 - 0.51616519j,
                -0.4936891 + 0.36476117j,
                0.70397844 + 0.05332291j,
                -0.56648961 - 0.50687223j,
                0.15372065 + 0.74444603j,
                0.30751808 - 0.63550106j,
                -0.5633934 + 0.24365451j,
                0.47550633 + 0.20903766j,
                -0.0985528 - 0.46673545j,
            ],
            [
                0.52924739 + 0.47782492j,
                0.08685642 - 0.77153733j,
                -0.73794122 + 0.54522635j,
                1.05227097 + 0.07970435j,
                -0.8467597 - 0.7576467j,
                0.22977376 + 1.11275987j,
                0.4596623 - 0.94991449j,
                -0.84213164 + 0.3642023j,
                0.71076254 + 0.3124588j,
                -0.14731169 - 0.69765229j,
            ],
            [
                -0.49897731 - 0.45049592j,
                -0.0818887 + 0.72740958j,
                0.69573498 - 0.51404236j,
                -0.99208678 - 0.0751457j,
                0.79832963 + 0.71431342j,
                -0.21663195 - 1.04911604j,
                -0.43337211 + 0.89558454j,
                0.79396628 - 0.3433719j,
                -0.67011078 - 0.29458785j,
                0.13888626 + 0.65775036j,
            ],
            [
                -0.2717424 - 0.2453395j,
                -0.04459648 + 0.39614632j,
                0.37889637 - 0.2799468j,
                -0.54028918 - 0.04092425j,
                0.43476929 + 0.38901417j,
                -0.11797748 - 0.57134724j,
                -0.23601389 + 0.48773418j,
                0.43239301 - 0.18699989j,
                -0.36494147 - 0.16043216j,
                0.07563728 + 0.35821j,
            ],
        ]
    )

    np.testing.assert_almost_equal(dmd.predict(X.T), expected, decimal=6)


def test_predict_exact():
    dmd = DMD(exact=True)
    expected = np.load("tests/test_datasets/input_sample_predict_exact.npy")

    np.testing.assert_almost_equal(
        dmd.fit(sample_data).predict(sample_data[:, 20:40]), expected, decimal=6
    )


def test_predict_nexact():
    dmd = DMD(exact=False)
    expected = np.load("tests/test_datasets/input_sample_predict_nexact.npy")

    np.testing.assert_almost_equal(
        dmd.fit(sample_data).predict(sample_data[:, 10:30]), expected, decimal=6
    )


def test_advanced_snapshot_parameter():
    dmd = DMD(svd_rank=0.99)
    dmd.fit(sample_data)

    dmd2 = DMD(svd_rank=0.99, opt=-1)
    dmd2.fit(sample_data)

    np.testing.assert_almost_equal(
        dmd2.reconstructed_data.real, dmd.reconstructed_data.real, decimal=6
    )


def test_sorted_eigs_default():
    dmd = DMD()
    assert dmd.operator._sorted_eigs == False


def test_sorted_eigs_set_real():
    dmd = DMD(sorted_eigs="real")
    assert dmd.operator._sorted_eigs == "real"


def test_sorted_eigs_abs_right_eigs():
    dmd = DMD(svd_rank=20, sorted_eigs="abs")
    dmd.fit(sample_data)

    dmd2 = DMD(svd_rank=20)
    dmd2.fit(sample_data)

    assert len(dmd.eigs) == len(dmd2.eigs)
    assert set(dmd.eigs) == set(dmd2.eigs)

    previous = dmd.eigs[0]
    for eig in dmd.eigs[1:]:
        assert abs(previous) <= abs(eig)
        previous = eig


def test_sorted_eigs_abs_right_eigenvectors():
    dmd = DMD(svd_rank=20, sorted_eigs="abs")
    dmd.fit(sample_data)

    dmd2 = DMD(svd_rank=20)
    dmd2.fit(sample_data)

    for idx, eig in enumerate(dmd2.eigs):
        eigenvector = dmd2.operator.eigenvectors.T[idx]
        for idx_new, eig_new in enumerate(dmd.eigs):
            if eig_new == eig:
                assert all(dmd.operator.eigenvectors.T[idx_new] == eigenvector)
                break


def test_sorted_eigs_abs_right_modes():
    dmd = DMD(svd_rank=20, sorted_eigs="abs")
    dmd.fit(sample_data)

    dmd2 = DMD(svd_rank=20)
    dmd2.fit(sample_data)

    for idx, eig in enumerate(dmd2.eigs):
        mode = dmd2.modes.T[idx]
        for idx_new, eig_new in enumerate(dmd.eigs):
            if eig_new == eig:
                np.testing.assert_almost_equal(
                    dmd.modes.T[idx_new], mode, decimal=6
                )
                break


def test_sorted_eigs_real_right_eigs():
    dmd = DMD(svd_rank=20, sorted_eigs="real")
    dmd.fit(sample_data)

    dmd2 = DMD(svd_rank=20)
    dmd2.fit(sample_data)

    assert len(dmd.eigs) == len(dmd2.eigs)
    assert set(dmd.eigs) == set(dmd2.eigs)

    previous = complex(dmd.eigs[0])
    for eig in dmd.eigs[1:]:
        x = complex(eig)
        assert x.real > previous.real or (
            x.real == previous.real and x.imag >= previous.imag
        )
        previous = x


def test_sorted_eigs_real_right_eigenvectors():
    dmd = DMD(svd_rank=20, sorted_eigs="real")
    dmd.fit(sample_data)

    dmd2 = DMD(svd_rank=20)
    dmd2.fit(sample_data)

    for idx, eig in enumerate(dmd2.eigs):
        eigenvector = dmd2.operator.eigenvectors.T[idx]
        for idx_new, eig_new in enumerate(dmd.eigs):
            if eig_new == eig:
                assert all(dmd.operator.eigenvectors.T[idx_new] == eigenvector)
                break


def test_sorted_eigs_real_right_modes():
    dmd = DMD(svd_rank=20, sorted_eigs="real")
    dmd.fit(sample_data)

    dmd2 = DMD(svd_rank=20)
    dmd2.fit(sample_data)

    for idx, eig in enumerate(dmd2.eigs):
        mode = dmd2.modes.T[idx]
        for idx_new, eig_new in enumerate(dmd.eigs):
            if eig_new == eig:
                np.testing.assert_almost_equal(
                    dmd.modes.T[idx_new], mode, decimal=6
                )
                break


def test_sorted_eigs_dynamics():
    dmd = DMD(svd_rank=20, sorted_eigs="abs")
    dmd.fit(sample_data)

    dmd2 = DMD(svd_rank=20)
    dmd2.fit(sample_data)

    for idx, eig in enumerate(dmd2.eigs):
        dynamic = dmd2.dynamics[idx]
        for idx_new, eig_new in enumerate(dmd.eigs):
            if eig_new == eig:
                np.testing.assert_almost_equal(
                    dmd.dynamics[idx_new], dynamic, decimal=6
                )
                break


def test_sorted_eigs_frequency():
    dmd = DMD(svd_rank=20, sorted_eigs="abs")
    dmd.fit(sample_data)

    dmd2 = DMD(svd_rank=20)
    dmd2.fit(sample_data)

    for idx, eig in enumerate(dmd2.eigs):
        frq = dmd2.frequency[idx]
        for idx_new, eig_new in enumerate(dmd.eigs):
            if eig_new == eig:
                np.testing.assert_almost_equal(
                    dmd.frequency[idx_new], frq, decimal=6
                )
                break


def test_sorted_eigs_amplitudes():
    dmd = DMD(svd_rank=20, sorted_eigs="abs")
    dmd.fit(sample_data)

    dmd2 = DMD(svd_rank=20)
    dmd2.fit(sample_data)

    for idx, eig in enumerate(dmd2.eigs):
        amp = dmd2.amplitudes[idx]
        for idx_new, eig_new in enumerate(dmd.eigs):
            if eig_new == eig:
                np.testing.assert_almost_equal(
                    dmd.amplitudes[idx_new], amp, decimal=6
                )
                break


def test_save():
    dmd = DMD(svd_rank=-1)
    dmd.fit(X=sample_data)
    dmd.save("pydmd.test")


def test_load():
    dmd = DMD(svd_rank=-1)
    dmd.fit(X=sample_data)
    dmd.save("pydmd.test2")
    loaded_dmd = DMD.load("pydmd.test2")
    np.testing.assert_array_equal(
        dmd.reconstructed_data, loaded_dmd.reconstructed_data
    )


def test_load():
    dmd = DMD(svd_rank=-1)
    dmd.fit(X=sample_data)
    dmd.save("pydmd.test2")
    loaded_dmd = DMD.load("pydmd.test2")
    assert isinstance(loaded_dmd, DMD)


def test_get_bitmask_default():
    dmd = DMD(svd_rank=10)
    dmd.fit(X=sample_data)
    assert np.all(dmd.modes_activation_bitmask == True)


def test_set_bitmask():
    dmd = DMD(svd_rank=3)
    dmd.fit(X=sample_data)

    new_bitmask = np.full(len(dmd.amplitudes), True, dtype=bool)
    new_bitmask[[0]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.modes_activation_bitmask[0] == False
    assert np.all(dmd.modes_activation_bitmask[1:] == True)


def test_not_fitted_get_bitmask_raises():
    dmd = DMD(svd_rank=3)
    with raises(RuntimeError):
        print(dmd.modes_activation_bitmask)


def test_not_fitted_set_bitmask_raises():
    dmd = DMD(svd_rank=3)
    with raises(RuntimeError):
        dmd.modes_activation_bitmask = np.full(3, True, dtype=bool)


def test_raise_wrong_dtype_bitmask():
    dmd = DMD(svd_rank=3)
    dmd.fit(X=sample_data)
    with raises(RuntimeError):
        dmd.modes_activation_bitmask = np.full(3, 0.1)


def test_fitted():
    dmd = DMD(svd_rank=3)
    assert not dmd.fitted
    dmd.fit(X=sample_data)
    assert dmd.fitted


def test_bitmask_amplitudes():
    dmd = DMD(svd_rank=10)
    dmd.fit(X=sample_data)

    old_n_amplitudes = dmd.amplitudes.shape[0]
    retained_amplitudes = np.delete(dmd.amplitudes, [0, -1])

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.amplitudes.shape[0] == old_n_amplitudes - 2
    np.testing.assert_almost_equal(dmd.amplitudes, retained_amplitudes)


def test_bitmask_eigs():
    dmd = DMD(svd_rank=10)
    dmd.fit(X=sample_data)

    old_n_eigs = dmd.eigs.shape[0]
    retained_eigs = np.delete(dmd.eigs, [0, -1])

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.eigs.shape[0] == old_n_eigs - 2
    np.testing.assert_almost_equal(dmd.eigs, retained_eigs)


def test_bitmask_modes():
    dmd = DMD(svd_rank=10)
    dmd.fit(X=sample_data)

    old_n_modes = dmd.modes.shape[1]
    retained_modes = np.delete(dmd.modes, [0, -1], axis=1)

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.modes.shape[1] == old_n_modes - 2
    np.testing.assert_almost_equal(dmd.modes, retained_modes)


def test_second_fit():
    dmd = DMD(svd_rank=-1)
    dmd.fit(X=sample_data)
    modes = dmd.modes
    proxy1 = dmd._modes_activation_bitmask_proxy

    dmd.fit(X=sample_data + 1)
    modes2 = dmd.modes
    proxy2 = dmd._modes_activation_bitmask_proxy

    assert proxy1 != proxy2
    with raises(AssertionError):
        np.testing.assert_allclose(modes, modes2)


def test_reconstructed_data_with_bitmask():
    dmd = DMD(svd_rank=10)
    dmd.fit(X=sample_data)

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    dmd.reconstructed_data
    assert True


def test_getitem_modes():
    dmd = DMD(svd_rank=-1)
    dmd.fit(X=sample_data)
    old_n_modes = dmd.modes.shape[1]

    assert dmd[[0, -1]].modes.shape[1] == 2
    np.testing.assert_almost_equal(dmd[[0, -1]].modes, dmd.modes[:, [0, -1]])

    assert dmd.modes.shape[1] == old_n_modes

    assert dmd[1::2].modes.shape[1] == old_n_modes // 2
    np.testing.assert_almost_equal(dmd[1::2].modes, dmd.modes[:, 1::2])

    assert dmd.modes.shape[1] == old_n_modes

    assert dmd[[1, 3]].modes.shape[1] == 2
    np.testing.assert_almost_equal(dmd[[1, 3]].modes, dmd.modes[:, [1, 3]])

    assert dmd.modes.shape[1] == old_n_modes

    assert dmd[2].modes.shape[1] == 1
    np.testing.assert_almost_equal(np.squeeze(dmd[2].modes), dmd.modes[:, 2])

    assert dmd.modes.shape[1] == old_n_modes


def test_getitem_raises():
    dmd = DMD(svd_rank=-1)
    dmd.fit(X=sample_data)

    with raises(ValueError):
        dmd[[0, 1, 1, 0, 1]]
    with raises(ValueError):
        dmd[[True, True, False, True]]
    with raises(ValueError):
        dmd[1.0]


def test_correct_amplitudes():
    dmd = DMD(svd_rank=-1)
    dmd.fit(X=sample_data)
    np.testing.assert_array_almost_equal(dmd.amplitudes, dmd._b)



"""
Module for the parametric Dynamic Mode Decomposition.

References:
- A Dynamic Mode Decomposition Extension for the Forecasting of Parametric 
Dynamical Systems, F. Andreuzzi, N. Demo, G. Rozza, 2023, SIAM Journal on 
Applied Dynamical Systems
"""

import pickle

import numpy as np


# roll by one position the shape of X. if X.shape == (a,b,c), the returned
# NumPy array's shape is (b,c,a)
def back_roll_shape(X):
    """
    Roll by one position the shape of `X`. if `X.shape == (a,b,c)`, the returned
    NumPy array's shape is `(b,c,a)`.
    """
    return np.swapaxes(np.swapaxes(X, 0, 1), 1, 2)


def roll_shape(X):
    """
    Roll by one position the shape of `X`. if `X.shape == (a,b,c)`, the returned
    NumPy array's shape is `(c,a,b)`.
    """
    return np.swapaxes(np.swapaxes(X, 0, 2), 1, 2)


class ParametricDMD:
    """
    Implementation of the parametric Dynamic Mode Decomposition proposed in
    arXiv:2110.09155v1. Both the *monolithic* and *partitioned* methods are
    available, see the documentation of the parameter `dmd` for more details.

    :param dmd: Instance(s) of :class:`dmdbase.DMDBase`, used by the
        paramtric DMD for the prediction of future spatial modal coefficients.
        If `dmd` is a `list` the *partitioned* approach is selected, in this
        case the number of parameters in the training set should be equal to
        the number of DMD instances provided. If `dmd` is not a list, we employ
        the monolithic approach.
    :type dmd: DMDBase or list
    :param spatial_pod: Instance of an object usable for the generation of a
        ROM of the dataset (see for instance the class
        `POD <https://mathlab.github.io/EZyRB/pod.html>`_ from the Python
        library `EZyRB <https://github.com/mathLab/EZyRB>`_).
    :param approximation: An interpolator following the standard
        learning-prediction pattern (`fit()` -> `predict()`). For some
        convenient wrappers see those implemented in
        `EZyRB <https://github.com/mathLab/EZyRB>`_).
    :param bool light: Whether this instance should be light or not. A light
        instance uses less memory since it caches a smaller number of resources.
        Setting `light=True` might invalidate several properties (see also
        :meth:`training_modal_coefficients`).
    :param dmd_fit_args: Positional arguments to be passed to the `fit` method
        of the given DMD instance.
    :param dmd_fit_kwargs: Keyword arguments to be passed to the `fit` method
        of the given DMD instance.
    """

    def __init__(
        self,
        dmd,
        spatial_pod,
        approximation,
        light=False,
        dmd_fit_args=None,
        dmd_fit_kwargs=None,
    ):
        self._dmd = dmd
        self._spatial_pod = spatial_pod
        self._approximation = approximation

        if dmd_fit_args is None:
            dmd_fit_args = tuple()
        if not isinstance(dmd_fit_args, (list, tuple)):
            raise TypeError("Expected list, tuple or None for dmd_fit_args")
        self._dmd_fit_args = dmd_fit_args
        if dmd_fit_kwargs is None:
            dmd_fit_kwargs = {}
        if not isinstance(dmd_fit_kwargs, dict):
            raise TypeError("Expected dict or None for dmd_fit_kwargs")
        self._dmd_fit_kwargs = dmd_fit_kwargs

        self._training_parameters = None
        self._parameters = None
        self._ntrain = None
        self._time_instants = None
        self._space_dim = None
        self._light = light
        self._training_modal_coefficients = None

    @property
    def is_partitioned(self):
        """
        Return `True` if this instance is partitioned, `False` if it is
        monolithic.

        :type: bool
        """
        return self._dmd is not None and isinstance(self._dmd, list)

    @property
    def _reference_dmd(self):
        """
        An object used as a reference for several properties like
        :func:`dmd_time` and :func:`dmd_timesteps`. If this instance is
        monolithic the returned value is `self._dmd`, otherwise it is the first
        item of the list `self._dmd`.

        :return: The object used as a reference.
        :rtype: pydmd.DMDBase
        """
        if self.is_partitioned:
            return self._dmd[0]
        return self._dmd

    @property
    def dmd_time(self):
        """
        The time dictionary used by the reference DMD instance (see also
        :func:`_reference_dmd`). Note that when you set this attribute the
        value is set only for the reference DMD (see :func:`_reference_dmd`),
        however when :func:`_predict_modal_coefficients` is called the values
        of all DMDs become consistent.

        :getter: Return the time dictionary used by the reference DMD instance.
        :setter: Set the given time dictionary in the field `dmd_time` for all
            DMD instances.
        :type: pydmd.dmdbase.DMDTimeDict
        """
        return self._reference_dmd.dmd_time

    @dmd_time.setter
    def dmd_time(self, value):
        self._reference_dmd.dmd_time = value

    @property
    def dmd_timesteps(self):
        """
        The timesteps in the output of this instance, which coincides with the
        timesteps in the output of the reference of this instance (see
        :func:`_reference_dmd`).

        :return: The timesteps in the output of this instance.
        :rtype: list
        """
        return self._reference_dmd.dmd_timesteps

    @property
    def original_time(self):
        """
        The original time dictionary used by this instance, which coincides
        with the original dictionary used by the reference of this instance
        (see :func:`_reference_dmd`).

        :return: The original time dictionary used by this instance.
        :rtype: dict
        """
        return self._reference_dmd.original_time

    @property
    def original_timesteps(self):
        """
        The original timesteps in the input fed to this instance, which
        coincides with the original timesteps in the input fed to the reference
        of this instance (see :func:`_reference_dmd`).

        :return: The original timesteps in the input fed to this instance.
        :rtype: list
        """
        return self._reference_dmd.original_timesteps

    @property
    def training_parameters(self):
        """
        The original parameters passed when `self.fit` was called, represented
        as a 2D array (the index of the parameter vary along the first
        dimension).

        :type: numpy.ndarray
        """
        return self._training_parameters

    def _set_training_parameters(self, params):
        """
        Set the value of `self._original_parameters`, while checking that the
        value provided is a 2D array.

        :param numpy.ndarray: A 2D array which contains the original
            parameters.
        """
        if isinstance(params, list):
            params = np.array(params)
        if params.ndim == 1:
            params = params[:, None]
        if params.ndim > 2:
            raise ValueError("Parameters must be stored in 2D arrays.")

        self._training_parameters = params

    @property
    def parameters(self):
        """
        The new parameters to be used in `reconstructed_data`, represented
        as a 2D array (the index of the parameter vary along the first
        dimension). For, instance, the following feeds a set of four 3D
        parameters to `ParametricDMD`:

            >>> from pydmd import ParametricDMD
            >>> pdmd = ParametricDMD(...)
            >>> pdmd.fit(...)
            >>> p0 = [0.1, 0.2, 0.1]
            >>> p1 = [0.1, 0.2, 0.3],
            >>> p2 = [0.2, 0.2, 0.2],
            >>> p3 = [0.1, 0.2, 0.2]
            >>> pdmd.parameters = np.array([p0,p1,p2,p3])

        Therefore, when we collect the results from `reconstructed_data`:

            >>> result = pdmd.reconstructed_data
            >>> # reconstruction corresponding to p0
            >>> rec_p0 = result[0]
            >>> # reconstruction corresponding to p1
            >>> rec_p1 = result[1]
            >>> ...

        :getter: Return the current parameters.
        :setter: Change the current parameters.
        :type: numpy.ndarray
        """
        return self._parameters if hasattr(self, "_parameters") else None

    @parameters.setter
    def parameters(self, value):
        if isinstance(value, list):
            value = np.array(value)
        if value.ndim == 1:
            value = value[:, None]
        elif value.ndim > 2:
            raise ValueError("Parameters must be stored in 2D arrays.")

        self._parameters = value

    def _arrange_parametric_snapshots(self, X):
        """
        Arrange the given parametric snapshots (see :func:`fit` for an overview
        of the shape of `X`) into a 2D matrix such that the shape is distributed
        as follows:

        - 0: Space;
        - 1: Time/Parameter.

        Time varies faster than the parameter along the columns of the matrix.

        An overview of the shape of the resulting matrix:

         .. math::

            M = \\begin{bmatrix}
                    x_1(t_1,\\mu_1) & \dots & x_1(t_n,\\mu_1) & x_1(t_1,\\mu_1)
                        & \dots & x_1(t_{n-1},\\mu_k) & x_1(t_n,\\mu_k)\\\\
                    \\vdots & \\dots & \\vdots & \\vdots & \\dots & \\vdots
                        & \\dots\\\\
                    x_m(t_1,\\mu_1) & \dots & x_m(t_n,\\mu_1) & x_m(t_1,\\mu_1)
                        & \dots & x_m(t_{n-1},\\mu_k) & x_m(t_n,\\mu_k)
                \\end{bmatrix}

        :math:`x(t, \mu) \in \mathbb{R}^m` is the functon which represents the
        parametric system at time :math:`t` with the parameter :math:`\\mu`.

        :param X: Parametric snapshots (distribition of axes like in
            :func:`fit`).
        :type X: numpy.ndarray
        :return: Parametric snapshots arranged in a 2D matrix like explained
            above.
        :rtype: numpy.ndarray
        """
        # swap parameters dimension and space dimension
        X = np.swapaxes(X, 0, 1)
        return X.reshape((X.shape[0], -1), order="C")

    def _compute_training_modal_coefficients(self, space_timemu):
        """
        Compute the POD modal coefficient from the given matrix, and put
        the resulting coefficients (along with their time evolution in matrix
        form) into a list.

        In symbols, from the given matrix :math:`X^x_{t,\mu} \in
        \mathbb{R}^{m \\times nk}` we compute the modal
        coefficients corresponding to its columns. At this point we have
        something like this:

        .. math::

            \\widetilde{X}^s_{t,\mu} = \\begin{bmatrix}
                    \\widetilde{x}_1(t_1,\\mu_1), & \dots &
                        \\widetilde{x}_1(t_n,\\mu_1), &
                        \\widetilde{x}_1(t_1,\\mu_1), & \dots &
                        \\widetilde{x}_1(t_{n-1},\\mu_k), &
                        \\widetilde{x}_1(t_n,\\mu_k)\\\\
                    \\vdots & \\dots & \\vdots & \\vdots & \\dots & \\vdots &
                        \\dots\\\\
                    \\widetilde{x}_p(t_1,\\mu_1), & \dots & x_p(t_n,\\mu_1) &
                        \\widetilde{x}_p(t_1,\\mu_1), & \dots &
                        \\widetilde{x}_p(t_{n-1},\\mu_k), &
                        \\widetilde{x}_p(t_n,\\mu_k)
                \\end{bmatrix} \in \mathbb{R}^{p \\times nk}

        Detecting the sub-matrices corresponding to the time evolution of the
        POD modal coefficients corresponding to a particular realization of the
        system for some parameter :math:`\\mu_i`, we may rewrite this matrix as
        follows:

        .. math::

            \\widetilde{X}^s_{t,\mu} = \\begin{bmatrix}
                    \\widetilde{X}_{\\mu_1}, & \dots & \\widetilde{X}_{\\mu_1}
            \\end{bmatrix}

        The returned list contains the matrices
        :math:`\\widetilde{X}_{\\mu_i} \in \\mathbb{p \\times n}`.

        :param space_timemu: A matrix containing parametric/time snapshots like
            the matrix returned by :func:`_arrange_parametric_snapshots`. The
            input size should be `p x nk` where `p` is the dimensionality of
            the full-dimensional space, `k` is the number of training parameters
            and `n` is the number of time instants used for the training.
        :type space_timemu: numpy.ndarray
        :return: A list of `k` matrices. Each matrix has shape `r x n` where `r`
            is the dimensionality of the reduced POD space, and `n`, `k` are the
            same of the parameter `space_timemu`.
        :rtype: list
        """

        spatial_modal_coefficients = self._spatial_pod.fit(space_timemu).reduce(
            space_timemu
        )
        return np.split(spatial_modal_coefficients, self._ntrain, axis=1)

    def fit(self, X, training_parameters):
        """
        Compute the parametric Dynamic Modes Decomposition from the input data
        stored in the array `X`. The shape of the parameter `X` must be
        used as follows:

        - 0: Training parameters;
        - 1: Space;
        - 2: Training time instants.

        The parameter `training_parameters` contains the list of training
        parameters corresponding to the training datasets in `X`. For instance,
        `training_parameters[0]` is the parameter which generated the dataset
        in `X[0]`. For this reason `len(training_parameters)` should be equal
        to `X.shape[0]`.

        :param numpy.ndarray X: Training snapshots of the parametric system,
            observed for two or more parameters and in multiple time instants.
        :param numpy.ndarray training_parameters: Training parameters
            corresponding to the snapshots in `X`.
        """

        if X.shape[0] != len(training_parameters):
            raise ValueError(
                "Unexpected number of snapshots for the given"
                "parameters. Received {} parameters, and {} snapshots".format(
                    len(training_parameters), X.shape[0]
                )
            )

        # we store these values for faster access
        self._ntrain, self._space_dim, self._time_instants = X.shape
        if self.is_partitioned and self._ntrain != len(self._dmd):
            raise ValueError(
                "Invalid number of DMD instances provided: "
                "expected n_train={}, got {}".format(
                    self._ntrain, len(self._dmd)
                )
            )

        # store the training parameters: they will be used in
        # `reconstructed_data`
        self._set_training_parameters(training_parameters)

        # arrange the parametric snapshots in a convenient way to perform POD
        space_timemu = self._arrange_parametric_snapshots(X)

        # obtain POD modal coefficients from the training set
        training_modal_coefficients = self._compute_training_modal_coefficients(
            space_timemu
        )

        if not self._light:
            self._training_modal_coefficients = np.array(
                training_modal_coefficients
            )

        # fit DMD(s) with POD modal coefficients
        if self.is_partitioned:
            # partitioned parametric DMD
            for dmd, data in zip(self._dmd, training_modal_coefficients):
                dmd.fit(data, *self._dmd_fit_args, **self._dmd_fit_kwargs)
        else:
            spacemu_time = np.vstack(training_modal_coefficients)
            self._dmd.fit(
                spacemu_time, *self._dmd_fit_args, **self._dmd_fit_kwargs
            )

    # ------------------------------------------------------------
    # getter properties for intermediate values of the computation

    @property
    def training_modal_coefficients(self):
        """
        Modal coefficients of the input dataset. Since this is cached after
        calls to :func:`fit` this property needs to be called after :func:`fit`,
        and `light` should be set to `False` in the constructor of the class.

        The tensor returned has the following shape:

        - 0: Training parameters;
        - 1: Dimensionality of the POD sub-space;
        - 2: Time.
        """
        if self._light:
            raise RuntimeError(
                """Light instances do not cache the property
`training_modal_coefficients`."""
            )

        if self._training_modal_coefficients is None:
            raise RuntimeError(
                """
Property not available now, did you call `fit()`?"""
            )

        return self._training_modal_coefficients

    @property
    def forecasted_modal_coefficients(self):
        """
        Modal coefficients forecasted for the input parameters.

        The tensor returned has the following shape:

        - 0: Training parameters;
        - 1: Dimensionality of the POD sub-space;
        - 2: Time.
        """
        forecasted = self._predict_modal_coefficients()
        return forecasted.reshape((self._ntrain, -1, forecasted.shape[1]))

    @property
    def interpolated_modal_coefficients(self):
        """
        Modal coefficients forecasted and then interpolated for the untested
        parameters.

        The tensor returned has the following shape:

        - 0: Parameters;
        - 1: Dimensionality of the POD sub-space;
        - 2: Time.
        """
        forecasted = self._predict_modal_coefficients()
        return self._interpolate_missing_modal_coefficients(forecasted)

    # ------------------------------------------------------------

    def _predict_modal_coefficients(self):
        """
        Predict future spatial modal coefficients in the time instants in
        `dmd_time`.

        :return: Predicted spatial modal coefficients. Shape: `rk x n` (`r`:
            dimensionality of POD subspace, `k`: number of training parameters,
            `n`: number of snapshots).
        :rtype: numpy.ndarray
        """
        if self.is_partitioned:
            for dmd in self._dmd:
                # we want to "bound" this DMD objects' dmd_time
                dmd.dmd_time = self._reference_dmd.dmd_time
            return np.vstack(
                list(map(lambda dmd: dmd.reconstructed_data, self._dmd))
            )
        return self._dmd.reconstructed_data

    def _interpolate_missing_modal_coefficients(
        self, forecasted_modal_coefficients
    ):
        """
        Interpolate spatial modal coefficients for the (untested) parameters
        stored in `parameters`. The interpolation uses the interpolator
        provided in the constructor of this instance.

        The returned value is a 3D tensor, its shape is used as follows:

        - 0: Parameters;
        - 1: Reduced POD space;
        - 2: Time.

        :param numpy.ndarray forecasted_modal_coefficients: An array of spatial
            modal coefficients for tested parameters. The shape is used like in
            the matrix returned by :func:`_predict_modal_coefficients`.
        :return: An array of (interpolated) spatial modal coefficients for
            untested parameters.
        :rtype: numpy.ndarray
        """

        if self.parameters is None or len(self.parameters) == 0:
            raise ValueError(
                """
Unknown parameters not found. Did you set `ParametricDMD.parameters`?"""
            )

        approx = self._approximation
        forecasted_modal_coefficients = forecasted_modal_coefficients.reshape(
            (self._ntrain, -1, forecasted_modal_coefficients.shape[1]),
            order="C",
        )

        def interpolate_future_pod_coefficients(time_slice):
            approx.fit(self.training_parameters, time_slice)
            return approx.predict(self.parameters)

        return np.dstack(
            [
                interpolate_future_pod_coefficients(time_slice)[..., None]
                for time_slice in roll_shape(forecasted_modal_coefficients)
            ]
        )

    @property
    def reconstructed_data(self):
        """
        Get the reconstructed data, for the time instants specified in
        `dmd_time`, and the parameters stored in `parameters`.

        The shape of the returned data is distributed as follows:

        - 0: Parameters;
        - 1: Space;
        - 2: Time.

        :return: Snapshots predicted/interpolated using parametric DMD and the
            given method for ROM.
        :rtype: numpy.ndarray
        """
        forecasted_modal_coefficients = self._predict_modal_coefficients()
        interpolated_modal_coefficients = (
            self._interpolate_missing_modal_coefficients(
                forecasted_modal_coefficients
            )
        )

        return np.apply_along_axis(
            self._spatial_pod.expand, 1, interpolated_modal_coefficients
        )

    def save(self, fname):
        """
        Save the object to `fname` using the pickle module.

        :param str fname: the name of file where the reduced order model will
            be saved.

        Example:

        >>> from pydmd import ParametricDMD
        >>> pdmd = ParametricDMD(...) #  Construct here the rom
        >>> pdmd.fit(...)
        >>> pdmd.save('pydmd.pdmd')
        """
        with open(fname, "wb") as output:
            pickle.dump(self, output, pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def load(fname):
        """
        Load the object from `fname` using the pickle module.

        :return: The `ReducedOrderModel` loaded

        Example:

        >>> from pydmd import ParametricDMD
        >>> pdmd = ParametricDMD.load('pydmd.pdmd')
        >>> print(pdmd.reconstructed_data)
        """
        with open(fname, "rb") as output:
            return pickle.load(output)



"""
Module for the Hankel alternative view of Koopman (HAVOK) analysis.

References:
- S. L. Brunton, B. W. Brunton, J. L. Proctor, E. Kaiser, and J. N. Kutz,
Chaos as an intermittently forced linear system, Nature Communications, 8
(2017), pp. 1-9.
- S. M. Hirsh, S. M. Ichinaga, S. L. Brunton, J. N. Kutz, and B. W. Brunton,
Structured time-delay models for dynamical systems with connections to
frenet-serret frame, Proceedings of the Royal Society A, 477
(2021). art. 20210097.
"""

import warnings
import numpy as np
import matplotlib.pyplot as plt

from matplotlib.gridspec import GridSpec
from scipy.signal import lsim, StateSpace
from scipy.stats import norm

from .bopdmd import BOPDMD
from .utils import compute_svd, differentiate


class HAVOK:
    """
    Hankel alternative view of Koopman (HAVOK) analysis.

    :param svd_rank: the rank for the truncation; if 0, the method computes the
        optimal rank and uses it for the truncation; if positive integer, the
        method uses the argument for the truncation; if float between 0 and 1,
        the rank is the number of the biggest singular values that are needed
        to reach the 'energy' specified by `svd_rank`; if -1, the method does
        not compute a truncation.
    :type svd_rank: int or float
    :param delays: the number of consecutive time-shifted copies of the
        data to use when building Hankel matrices. Note that if examining an
        n-dimensional data set, this means that the resulting Hankel matrix
        will contain n * `delays` rows.
    :type delays: int
    :param lag: the number of time steps between each time-shifted copy of
        data in the Hankel matrix. This means that each row of the Hankel
        matrix will be separated by a time-step of dt * `lag`.
    :type lag: int
    :param num_chaos: the number of forcing terms to use in the HAVOK model.
    :type num_chaos: int
    :param structured: whether to perform standard HAVOK or structured HAVOK
        (sHAVOK). If `True`, sHAVOK is performed, otherwise HAVOK is performed.
        Note that sHAVOK cannot be performed with a `BOPDMD` model.
    :type structured: bool
    :param lstsq: method used for computing the HAVOK operator if a DMD method
        is not provided. If True, least-squares is used, otherwise the pseudo-
        inverse is used. This parameter is ignored if `dmd` is provided.
    :type lstsq: bool
    :param dmd: DMD instance used to compute the HAVOK operator. If `None`,
        least-squares or the pseudo-inverse is used depending on `lstsq`.
    :type dmd: DMDBase
    """

    def __init__(
        self,
        svd_rank=0,
        delays=10,
        lag=1,
        num_chaos=1,
        structured=False,
        lstsq=True,
        dmd=None,
    ):
        self._svd_rank = svd_rank
        self._delays = delays
        self._lag = lag
        self._num_chaos = num_chaos
        self._structured = structured
        self._lstsq = lstsq
        self._dmd = dmd

        # Keep track of the original data and Hankel matrix.
        self._snapshots = None
        self._ho_snapshots = None
        self._time = None

        # Keep track of SVD information.
        self._singular_vecs = None
        self._singular_vals = None
        self._delay_embeddings = None

        # Keep track of the full HAVOK operator.
        self._havok_operator = None
        self._eigenvalues = None
        self._r = None

    @property
    def delays(self):
        """
        Get the number of delays used when building Hankel matrices.

        :return: the number of Hankel matrix delays.
        :rtype: int
        """
        return self._delays

    @property
    def lag(self):
        """
        Get the lag used when building Hankel matrices.

        :return: the number of time-steps used for Hankel matrix lag.
        :rtype: int
        """
        return self._lag

    @property
    def snapshots(self):
        """
        Get the input data (time-series or space-flattened).

        :return: the matrix that contains the original input data.
        :rtype: numpy.ndarray
        """
        if self._snapshots is None:
            raise ValueError("You need to call fit().")
        return np.squeeze(self._snapshots)

    @property
    def ho_snapshots(self):
        """
        Get the time-delay data matrix (i.e. the Hankel matrix).

        :return: the matrix that contains the time-delayed data.
        :rtype: numpy.ndarray
        """
        if self._ho_snapshots is None:
            raise ValueError("You need to call fit().")
        return self._ho_snapshots

    @property
    def time(self):
        """
        Get the times of the input data.

        :return: the vector that contains the times of the input data.
        :rtype: numpy.ndarray
        """
        if self._time is None:
            raise ValueError("You need to call fit().")
        return self._time

    @property
    def modes(self):
        """
        Get the U matrix from the SVD of the Hankel matrix. Note that the
        columns of this matrix are referred to as the eigen-time-delay modes.

        :return: matrix containing the eigen-time-delay modes.
        :rtype: numpy.ndarray
        """
        if self._singular_vecs is None:
            raise ValueError("You need to call fit().")
        return self._singular_vecs

    @property
    def singular_vals(self):
        """
        Get the singular value spectrum of the Hankel matrix.

        :return: the singular values of the Hankel matrix.
        :rtype: numpy.ndarray
        """
        if self._singular_vals is None:
            raise ValueError("You need to call fit().")
        return self._singular_vals

    @property
    def delay_embeddings(self):
        """
        Get all of the HAVOK embeddings (linear dynamics and forcing).
        Coordinates are stored as columns of the returned matrix.
        Note that this is the V matrix from the SVD of the Hankel matrix.

        :return: matrix containing all of the HAVOK embeddings.
        :rtype: numpy.ndarray
        """
        if self._delay_embeddings is None:
            raise ValueError("You need to call fit().")
        return self._delay_embeddings

    @property
    def linear_dynamics(self):
        """
        Get the HAVOK embeddings that are governed by linear dynamics.
        Coordinates are stored as columns of the returned matrix.

        :return: matrix containing the linear HAVOK embeddings.
        :rtype: numpy.ndarray
        """
        if self._delay_embeddings is None:
            raise ValueError("You need to call fit().")
        return self._delay_embeddings[:, : -self._num_chaos]

    @property
    def forcing(self):
        """
        Get the HAVOK embeddings that force the linear dynamics.
        Coordinates are stored as columns of the returned matrix.

        :return: matrix containing the chaotic forcing terms.
        :rtype: numpy.ndarray
        """
        if self._delay_embeddings is None:
            raise ValueError("You need to call fit().")
        return self._delay_embeddings[:, -self._num_chaos :]

    @property
    def operator(self):
        """
        Get the full HAVOK regression model,
        which contains A, B, and the bad fit.

        :return: the full HAVOK regression model.
        :rtype: numpy.ndarray
        """
        if self._havok_operator is None:
            raise ValueError("You need to call fit().")
        return self._havok_operator

    @property
    def A(self):
        """
        Get the matrix A in the HAVOK relationship dv/dt = Av + Bu, where v
        denotes the linear HAVOK embeddings and u denotes the forcing terms.

        :return: linear dynamics matrix A.
        :rtype: numpy.ndarray
        """
        if self._havok_operator is None:
            raise ValueError("You need to call fit().")
        return self._havok_operator[: -self._num_chaos, : -self._num_chaos]

    @property
    def B(self):
        """
        Get the matrix B in the HAVOK relationship dv/dt = Av + Bu, where v
        denotes the linear HAVOK embeddings and u denotes the forcing terms.

        :return: forcing dynamics matrix B.
        :rtype: numpy.ndarray
        """
        if self._havok_operator is None:
            raise ValueError("You need to call fit().")
        return self._havok_operator[: -self._num_chaos, -self._num_chaos :]

    @property
    def eigs(self):
        """
        Get the eigenvalues of the linear HAVOK operator A.

        :return: the eigenvalues of the operator A.
        :rtype: numpy.ndarray
        """
        if self._eigenvalues is None:
            raise ValueError("You need to call fit().")
        return self._eigenvalues

    @property
    def r(self):
        """
        Get the number of HAVOK embeddings utilized by the HAVOK model.
        Note that this is essentially the integer rank truncation used.

        :return: rank of the HAVOK model.
        :rtype: int
        """
        if self._r is None:
            raise ValueError("You need to call fit().")
        return self._r

    def hankel(self, X):
        """
        Given a data matrix X as a 1-D or 2-D numpy.ndarray, uses the `delays`
        and `lag` attributes to return the data as a 2-D Hankel matrix.

        :param X: (m,) or (n, m) array of data.
        :type X: numpy.ndarray
        :return: Hankel matrix of data.
        :rtype: numpy.ndarray
        """
        if not isinstance(X, np.ndarray) or X.ndim > 2:
            raise ValueError("Data must be a 1-D or 2-D numpy array.")
        if X.ndim == 1:
            X = X[None]
        n, m = X.shape

        # Check that the input data contains enough observations.
        m_min = self._lag * (self._delays - 1) + 1
        if m < m_min:
            raise ValueError(
                "Not enough snapshots provided for "
                f"{self._delays} delays and lag {self._lag}. "
                f"Please provide at least {m_min} snapshots."
            )

        Hm = m - ((self._delays - 1) * self._lag)
        H = np.empty((n * self._delays, Hm))
        for i in range(self._delays):
            H[i * n : (i + 1) * n] = X[:, i * self._lag : i * self._lag + Hm]

        return H

    def dehankel(self, H):
        """
        Given a Hankel matrix H as a 2-D numpy.ndarray, uses the `delays`
        and `lag` attributes to unravel the data in the Hankel matrix.

        :param H: 2-D Hankel matrix of data.
        :type H: numpy.ndarray
        :return: de-Hankeled (m,) or (n, m) array of data.
        :rtype: numpy.ndarray
        """
        if not isinstance(H, np.ndarray) or H.ndim != 2:
            raise ValueError("Data must be a 2-D numpy array.")

        Hn, Hm = H.shape
        n = int(Hn / self._delays)
        m = int(Hm + ((self._delays - 1) * self._lag))
        X = np.empty((n, m))
        for i in range(self._delays):
            X[:, i * self._lag : i * self._lag + Hm] = H[i * n : (i + 1) * n]

        return np.squeeze(X)

    def fit(self, X, t):
        """
        Perform the HAVOK analysis.

        :param X: the input snapshots.
        :type X: numpy.ndarray or iterable
        :param t: the input time vector or uniform time-step between snapshots.
        :type t: {numpy.ndarray, iterable} or {int, float}
        """

        # Confirm that delays, lag, and num_chaos are positive integers.
        for x in [self._delays, self._lag, self._num_chaos]:
            if not isinstance(x, int) or x < 1:
                raise ValueError(
                    "delays, lag, and num_chaos must be positive integers."
                )

        # Confirm that the input data is 1-D or 2-D.
        X = np.squeeze(np.array(X))
        if X.ndim > 2:
            raise ValueError("Input data must be a 1-D or 2-D array.")
        if X.ndim == 1:
            X = X[None]
        n_samples = X.shape[-1]

        # Compute the Hankel matrix.
        hankel_matrix = self.hankel(X)

        # Check the input time information and set the time vector.
        if isinstance(t, (int, float)):
            time = np.arange(n_samples) * t
        else:
            time = np.squeeze(np.array(t))

            # Throw error if the time vector is not 1-D or the correct length.
            if time.ndim != 1 or len(time) != n_samples:
                raise ValueError(
                    f"Please provide a 1-D array of {n_samples} time values."
                )

            # Generate warning if the times are not uniformly-spaced.
            if not np.allclose(time[1:] - time[:-1], time[1] - time[0]):
                warnings.warn(
                    "Input snapshots are unevenly-spaced in time. "
                    "Unexpected results may occur because of this."
                )

        # Set the time step - this is ignored if using BOP-DMD.
        dt = time[1] - time[0]

        # Perform structured HAVOK (sHAVOK).
        if self._structured:
            U, s, V = compute_svd(hankel_matrix[:, :-1], self._svd_rank)
            self._r = len(s)
            V2 = compute_svd(hankel_matrix[:, 1:], self._r)[-1]
            V_dot = (V2 - V) / dt

        # Perform standard HAVOK.
        else:
            U, s, V = compute_svd(hankel_matrix, self._svd_rank)
            self._r = len(s)
            V_dot = differentiate(V.T, dt).T

        # Generate an error if too few HAVOK embeddings are being used.
        if self._r < self._num_chaos + 1:
            raise ValueError(
                f"HAVOK is attempting to use r = {self._r} embeddings "
                f"when r should be at least {self._num_chaos + 1}. "
                "Try increasing the number of delays or providing "
                "a positive integer argument for svd_rank."
            )

        # Use lstsq or pinv to compute the HAVOK operator.
        if self._dmd is None:
            if self._lstsq:
                havok_operator = np.linalg.lstsq(V, V_dot, rcond=None)[0].T
            else:
                havok_operator = np.linalg.pinv(V).dot(V_dot).T

        # Use the provided DMDBase object to compute the operator.
        else:
            if isinstance(self._dmd, BOPDMD):
                self._dmd.fit(V.T, time[: len(V)])

                if self._structured:
                    warnings.warn(
                        "Structured HAVOK cannot be performed with BOP-DMD. "
                        "Performing normal HAVOK instead..."
                    )
            else:
                self._dmd.fit(V.T, V_dot.T)

            # Compute the full system matrix.
            havok_operator = np.linalg.multi_dot(
                [
                    self._dmd.modes,
                    np.diag(self._dmd.eigs),
                    np.linalg.pinv(self._dmd.modes),
                ]
            )

        # Set the input data information.
        self._snapshots = X
        self._ho_snapshots = hankel_matrix
        self._time = time

        # Set the SVD information.
        self._singular_vecs = U
        self._singular_vals = s
        self._delay_embeddings = V

        # Save the full HAVOK operator.
        self._havok_operator = havok_operator
        self._eigenvalues = np.linalg.eig(
            havok_operator[: -self._num_chaos, : -self._num_chaos]
        )[0]

        return self

    def predict(self, forcing, time, V0=0):
        """
        Use a custom forcing input to make system predictions.

        :param forcing: (m, `num_chaos`) array of forcing inputs.
        :type forcing: numpy.ndarray
        :param time: (m,) array that contains the times that correspond with
            the provided forcing inputs. These will also be the times at which
            system predictions are computed.
        :type time: numpy.ndarray
        :param V0: (`r` - `num_chaos`,) array that contains the initial
            condition of the linear dynamics. This array should contain the
            linear dynamics evaluated at the first time in the `time` array.
            If not provided, this initial condition is assumed to be the
            initial condition stored in this `HAVOK` model instance. If `V0`
            is an int, `V0` is assumed to be the index of the stored linear
            dynamics to use as an initial condition.
        :type V0: {numpy.ndarray, iterable} or int
        :return: system predictions evaluated at the times in `time`.
        :rtype: numpy.ndarray
        """
        if isinstance(V0, int):
            V0 = self.linear_dynamics[V0]
        else:
            V0 = np.array(V0)

        return np.squeeze(
            self._embeddings_to_original(
                self._compute_embeddings(forcing, time, V0)
            )
        )

    @property
    def reconstructed_embeddings(self):
        """
        Get the reconstructed time-delay embeddings.

        :return: the matrix that contains the reconstructed embeddings.
        :rtype: numpy.ndarray
        """
        return self._compute_embeddings(
            self.forcing,
            self._time[: len(self.forcing)],
            self.linear_dynamics[0],
        )

    @property
    def reconstructed_data(self):
        """
        Get the reconstructed data.

        :return: the matrix that contains the reconstructed snapshots.
        :rtype: numpy.ndarray
        """
        return np.squeeze(
            self._embeddings_to_original(self.reconstructed_embeddings)
        )

    def compute_threshold(
        self,
        forcing=0,
        p=0.01,
        bins=50,
        plot=False,
        plot_kwargs=None,
    ):
        """
        Use the distribution of forcing terms to determine a threshold at which
        the absolute value of the forcing is large enough to be considered
        "active". This method uses a histogram of the forcing signal values
        and a forcing event probability in order to estimate this threshold.

        :param forcing: (m,) array of forcing inputs to be thresholded.
            Alternatively, users may provide an integer, which will be used to
            index the stored forcing terms. By default, the first forcing term
            stored will be used.
        :type forcing: {numpy.ndarray, iterable} or int
        :param p: desired approximate probability that a forcing event occurs.
            Note that `p` must be a float between 0.0 and 1.0, and that smaller
            values of `p` will result in larger threshold values. If `p` is an
            integer instead, `p` will be used to index candidate thresholds
            that are located at the intersection of the forcing term histogram
            and a fitted Gaussian distribution.
        :type p: int or float
        :param bins: `bins` input to the `numpy.histogram` function.
        :type bins: int or sequence of scalars or str
        :param plot: whether or not to plot the computed histogram of forcing
            values and the computed threshold. A Gaussian distribution fitted
            to the computed histogram is also plotted if `plot=True`.
        :type plot: bool
        :param plot_kwargs: optional dictionary of plot parameters. Currently,
            one may set the figure size, the y-axis limits, and whether or not
            to use a semilogy scale.
        :type plot_kwargs: dict
        :return: active threshold for the absolute value of the forcing terms.
        :rtype: float
        """
        if isinstance(forcing, int):
            forcing = self.forcing[:, forcing]
        else:
            forcing = np.array(forcing)

        # Compute histogram of the forcing values.
        hy, hx = np.histogram(forcing, bins=bins, density=True)
        hx = 0.5 * (hx[:-1] + hx[1:])  # get bin centers
        hy /= hy.sum()

        # Fit a Gaussian to the forcing values.
        mu, std = norm.fit(forcing)
        gauss = norm.pdf(hx, mu, std)
        gauss /= np.sum(gauss)

        if isinstance(p, int):
            # Use Gaussian intersection.
            a = gauss - hy
            sgn = np.sign(a)
            ind_signchange = np.where(sgn(a[:-1]) * sgn(a[1:]) > 0)[0]
            thres_1 = np.abs(hx[ind_signchange])
            thres_2 = np.abs(hx[ind_signchange + 1])
            threshold_candidates = np.sort(0.5 * (thres_1 + thres_2))
            threshold = threshold_candidates[p]
        else:
            # Use probability p of switching.
            ind1 = np.where(np.cumsum(hy) > 0.5 * p)[0][0]
            ind2 = np.where(1 - np.cumsum(hy) < 0.5 * p)[0][0]
            threshold = 0.5 * (abs(hx[ind1]) + abs(hx[ind2]))

        if plot:
            # Set the plotting parameters first.
            if plot_kwargs is None:
                plot_kwargs = {}
            if "figsize" not in plot_kwargs:
                plot_kwargs["figsize"] = (5, 4)
            if "semilogy" not in plot_kwargs:
                plot_kwargs["semilogy"] = True

            # Plot the histogram, fitted Gaussian, and the threshold.
            plt.figure(figsize=plot_kwargs["figsize"])
            plt.plot(hx, hy, c="tab:red", label="Forcing", lw=2)
            plt.plot(hx, gauss, c="k", label="Gaussian", lw=2, ls="--")
            plt.axvline(x=threshold, lw=2, label="Threshold")
            plt.axvline(x=-threshold, lw=2)
            plt.xlabel("$v_r$")
            plt.ylabel("$p$", rotation=0)
            plt.legend()
            if plot_kwargs["semilogy"]:
                plt.semilogy()
            if "ylim" in plot_kwargs:
                plt.ylim(plot_kwargs["ylim"])
            plt.show()

        return threshold

    def plot_summary(
        self,
        num_plot=None,
        index_linear=(0, 1, 2),
        index_forcing=0,
        forcing_threshold=None,
        min_jump_dist=None,
        true_switch_indices=None,
        figsize=(20, 4),
        dpi=200,
        filename=None,
    ):
        """
        Generate a 5-element summarizing plot that contains the following:
        - the time-series used to apply HAVOK
        - the full linear operator, which contains A, B, and the bad fit
        - the first linear embedding term and the first forcing term
        - the HAVOK embeddings, along with active forcing times
        - the HAVOK reconstruction of the embeddings.

        :param num_plot: The number of time points to plot across all subplots.
            By default, all available data points are plotted.
        :type num_plot: int
        :param index_linear: Tuple of indices of the linear embeddings to be
            plotted. May contain either 2 or 3 valid indices. The final two
            subplots will be plotted in 2-D or 3-D depending on the number of
            indices provided.
        :type index_linear: iterable
        :param index_forcing: Index of the forcing term to be plotted. Note
            that this index refers to indices of the forcing term itself rather
            than the full matrix of time-delay embeddings. Hence if 0, the
            first forcing term will be plotted, and so on.
        :type index_forcing: int
        :param forcing_threshold: Threshold value at which the absolute value
            of the forcing signal is considered large enough to be "active".
        :type forcing_threshold: float
        :param min_jump_dist: The minimum number of indices used to separate
            distinct forcing events. Decreasing this parameter will lead to
            many short forcing events, while increasing this parameter will
            lead to fewer longer forcing events.
        :type min_jump_dist: int
        :param true_switch_indices: Optional vector that contains the indices
            at which true chaotic bursting occurs. If provided, true bursting
            times are plotted on top of the forcing term.
        :type true_switch_indices: numpy.ndarray or iterable
        :param figsize: Tuple in inches defining the figure size.
        :type figsize: tuple(int, int)
        :param dpi: Figure resolution.
        :type dpi: int
        :param filename: If specified, the plot is saved at `filename`.
        :type filename: str
        """
        if self._havok_operator is None:
            raise ValueError("You need to call fit().")

        # Plot as many time points as possible.
        if num_plot is None:
            num_plot = len(self._delay_embeddings)

        # Compute a threshold based on the distribution of all forcing terms.
        if forcing_threshold is None:
            forcing_threshold = self.compute_threshold(
                self.forcing[:, index_forcing]
            )

        # Use the time step to estimate a reasonable jump.
        if min_jump_dist is None:
            min_jump_dist = int(0.5 / (self._time[1] - self._time[0]))

        # Get index slices at which the forcing is considered active.
        forcing = self.forcing[:num_plot, index_forcing]
        active_indices = np.arange(num_plot)[
            np.abs(forcing) > forcing_threshold
        ]
        active_slices = self._get_index_slices(active_indices, min_jump_dist)

        fig = plt.figure(figsize=figsize, dpi=dpi)
        gs = GridSpec(2, 5, figure=fig)
        ax1 = fig.add_subplot(gs[:, 0])
        ax2 = fig.add_subplot(gs[:, 1])
        ax3 = fig.add_subplot(gs[0, 2])
        ax4 = fig.add_subplot(gs[1, 2])
        if len(index_linear) == 3:
            ax5 = fig.add_subplot(gs[:, 3], projection="3d")
            ax6 = fig.add_subplot(gs[:, 4], projection="3d")
        else:
            ax5 = fig.add_subplot(gs[:, 3])
            ax6 = fig.add_subplot(gs[:, 4])

        # (1) plot the time-series data (first coordinate).
        ax1.set_title("Time series")
        ax1.plot(self._time[:num_plot], self._snapshots[0, :num_plot], c="k")
        ax1.set_xlabel("Time")

        # (2) plot the HAVOK operator.
        ax2.set_title("Regression model")
        vmax = np.abs(self._havok_operator.real).max()
        fig.colorbar(
            ax2.imshow(
                self._havok_operator.real,
                vmax=vmax,
                vmin=-vmax,
                cmap="PuOr",
            ),
            fraction=0.046,
            pad=0.04,
        )
        a = len(self._havok_operator) - self._num_chaos - 0.5
        ax2.plot([a, a], [-0.5, a], c="k", lw=1)
        ax2.axhline(y=a, c="k", lw=1)
        ax2.set_xticks([])
        ax2.set_yticks([])

        # (3) plot the linear HAVOK embeddings (first coordinate).
        ax3.set_title("Linear dynamics")
        ax3.plot(
            self._time[:num_plot],
            self.linear_dynamics[:num_plot, 0],
            c="tab:blue",
        )
        ax3.set_xticks([])
        ax3.set_yticks([])

        # (4) plot the HAVOK forcing term with activation times.
        ax4.set_title("Forcing")
        ax4.plot(self._time[:num_plot], forcing, c="gray")
        ax4.set_xticks([])
        ax4.set_yticks([])

        for ind1, ind2 in active_slices:
            ax4.plot(self._time[ind1:ind2], forcing[ind1:ind2], c="r")

        if true_switch_indices is not None:
            # Remove indices that fall outside of the plotting range.
            outside_indices = np.where(true_switch_indices >= num_plot)[0]
            if len(outside_indices) > 0:
                true_switch_indices = true_switch_indices[: outside_indices[0]]
            ax4.plot(
                self._time[:num_plot][true_switch_indices],
                np.zeros(len(true_switch_indices)),
                "*",
                mec="k",
                mfc="y",
                ms=12,
            )

        # (5) plot the embedded attractor with activation.
        ax5.set_title("Embedded attractor")
        linear_dynamics = [
            self.linear_dynamics[:num_plot, i] for i in index_linear
        ]
        ax5.plot(*linear_dynamics, c="gray")
        for ind1, ind2 in active_slices:
            active_dynamics = [
                self.linear_dynamics[ind1:ind2, i] for i in index_linear
            ]
            ax5.plot(*active_dynamics, c="r")
        ax5.set_axis_off()

        # (6) plot the reconstructed attractor.
        ax6.set_title("Reconstructed attractor")
        reconstruction = [
            self.reconstructed_embeddings[:num_plot, i] for i in index_linear
        ]
        ax6.plot(*reconstruction, c="tab:blue")
        ax6.set_axis_off()
        plt.tight_layout(pad=0.1)

        # Save plot if filename is provided.
        if filename:
            plt.savefig(filename)
            plt.close(fig)
        else:
            plt.show()

    def _compute_embeddings(self, forcing, time, V0):
        """
        Helper function that uses the fitted HAVOK model to reconstruct the
        time-delay embeddings for a generic forcing term, set of times, and
        initial condition for the time-delay embeddings.
        """
        # Build a system with the following form:
        #   dx/dt = Ax + Bu
        #   y = Cx + Du
        C = np.eye(len(self.A))
        D = 0.0 * self.B
        havok_system = StateSpace(self.A, self.B, C, D)

        # Reconstruct the linear dynamics using the HAVOK system.
        embeddings = lsim(
            havok_system,
            U=forcing,
            T=time,
            X0=V0,
        )[1]

        return embeddings

    def _embeddings_to_original(self, V):
        """
        Helper function that uses SVD and Hankel parameter information stored
        in the HAVOK model to convert data in time-delay embedding space back
        to the space of the original input data.
        """
        U = self._singular_vecs[:, : V.shape[-1]]
        s = self._singular_vals[: V.shape[-1]]
        H = np.linalg.multi_dot([U, np.diag(s), V.conj().T])
        return self.dehankel(H)

    @staticmethod
    def _get_index_slices(x, min_jump_dist):
        """
        Helper function that, given an array x of indices at which to plot,
        computes and returns the beginning and ending index for each
        consecutive set of indices.

        :Example:
            >>> a = np.array([2, 3, 4, 5, 10, 11, 12, 25, 26, 28])
            >>> _get_index_slices(a, min_jump_dist=2)
            [(2, 5), (10, 12), (25, 28)]
        """
        # Get the locations within x where a significant jump occurs.
        jumps = x[1:] - x[:-1] > min_jump_dist
        jump_starts = np.insert(x[1:][jumps], 0, x[0])
        jump_ends = np.append(x[:-1][jumps], x[-1])
        index_slices = list(zip(jump_starts, jump_ends))

        return index_slices
